import sys

import numpy as np
from scipy.interpolate import interp1d
from scipy.integrate import quad

from solcore import constants, material
from solcore.absorption_calculator import adachi_alpha
from solcore.material_data import calculate_mobility
from solcore.structure import *
from .QWunit import QWunit

Epsi0 = constants.vacuum_permittivity
q = constants.q
pi = constants.pi
h = constants.h
kb = constants.kb
m0 = constants.electron_mass
vacuum_permittivity = constants.vacuum_permittivity
c = constants.c

# The default material is a set of minimum properties that are used for the layers unless otherwise stated
DefaultMaterial = material("GaAs")(T=293)
DefaultProperties = {'band_gap': DefaultMaterial.band_gap,  # J
                     'electron_affinity': DefaultMaterial.electron_affinity,  # J
                     'eff_mass_electron_Gamma': DefaultMaterial.eff_mass_electron_Gamma,  # relative to m0
                     'eff_mass_hh_z': DefaultMaterial.eff_mass_hh_z,  # relative to m0
                     'eff_mass_lh_z': DefaultMaterial.eff_mass_lh_z,  # relative to m0
                     'electron_mobility': calculate_mobility("GaAs", 0, 1),  # m2 V-1 s-1
                     'hole_mobility': calculate_mobility("GaAs", 1, 1),  # m2 V-1 s-1
                     'ni': DefaultMaterial.ni,  # m-3
                     'Nc': DefaultMaterial.Nc,  # m-3
                     'Nv': DefaultMaterial.Nv,  # m-3
                     'electron_minority_lifetime': 3e-6,  # s
                     'hole_minority_lifetime': 2.5e-7,  # s
                     'relative_permittivity': DefaultMaterial.relative_permittivity,  # relative to epsilon0
                     'electron_auger_recombination': 1e-42,  # m6 s-1,
                     'hole_auger_recombination': 1e-42,  # m6 s-1
                     # 'radiative_recombination': 7.2e-16,  # m3 s-1
                     'radiative_recombination': DefaultMaterial.radiative_recombination,  # m3 s-1
                     'Nd': 1,  # m-3
                     'Na': 1}  # m3


def CreateDeviceStructure(name=None, role='device', T=293, layers=None, comments='', repeat=1, substrate=DefaultMaterial,
                          reflection=None):
    """ Creates the structure of a device to be used in drift diffusion calculations.

    :param name: Name of the structure
    :param role: Role of the structure (eg. "device", "MQW", etc)
    :param T: Temperature
    :param layers: List containing all the layers of the structure. It can also be a Structure or a Junction.
    :param comments: A string with comments
    :param repeat: If the structure is actually something that is repeated many times, for example a quantum well
    :param substrate: A Solcore material defining the substrate of the structure.
    :param reflection: The reflection of the structure -> Not use from Solcore v4
    :return: A dictionary-like object with all the information related to the device
    
    """
    output = {}
    output['name'] = name
    output['role'] = role
    output['T'] = T
    output['numlayers'] = 0
    output['comments'] = comments
    output['repeat'] = repeat
    output['layers'] = []
    output['substrate'] = SolcoreMaterialToStr(substrate)
    output['reflection'] = reflection
    output['absorption'] = layers.absorbed if hasattr(layers, 'absorbed') else lambda x: 0
    output['sn'] = layers.sn if hasattr(layers, 'sn') else 1e6
    output['sp'] = layers.sp if hasattr(layers, 'sp') else 1e6

    AddLayers(output, layers)

    return output


def AddLayers(device, layers):
    """ Add layers to the structure

    :param device: The device structure
    :param layers: A list with the layers to add
    :return: None
    """
    for layer in layers:
        NewLayer = {}

        NewLayer['label'] = layer.role
        NewLayer['class'] = 'Layer'
        NewLayer['group'] = None
        NewLayer['numlayers'] = 1
        NewLayer['repeat'] = 1
        NewLayer['properties'] = GetLayerProperties(layer, device['T'])

        device['layers'].append(NewLayer)
        device['numlayers'] = device['numlayers'] + 1


def RemoveLayer(device, i):
    """ Remove a layer from the device structure

    :param device: the device structure
    :param i: the index of the layer to remove
    :return: None
    """
    del device['layers'][i]
    device['numlayers'] = device['numlayers'] - 1


def GetLayerProperties(layer, T):
    """ Given a layer, get all the properties of that layer and the material it is made of at the given temperature

    :param layer: The layer of interest
    :param T: The temperature
    :return: Dictionary with all the properties
    """
    # Get all the properties of the new layer. If they dont exist in the layer definition, the default values are used. 
    NewProperties = {}
    NewProperties['composition'] = SolcoreMaterialToStr(layer.material)
    NewProperties['width'] = layer.width

    for key in DefaultProperties.keys():
        try:
            NewProperties[key] = getattr(layer.material, key)
        except ValueError:
            NewProperties[key] = DefaultProperties[key]

    return NewProperties


def LoadAbsorption(layer, T, wavelengths, use_Adachi=False):
    """ Loads the absorption coefficient for the material of a given layer. It is only useful when solving QWs, since it is necessary to calculate an effective absorption coefficient before solving the opticalproperties of the structure using the BL, TMM or RCWA solvers.

    :param layer: The layer we want to get the absorption coefficient from.
    :param T: The temperature
    :param wavelengths: The wavelength we want the absorption at
    :param use_Adachi: If we should use the Adachi method to get the absorption coeficient.
    :return: A list with two columns, the wavelengths and the absorption coeficient.
    """

    if use_Adachi:
        try:
            # 0 = Energy, 1 = n, 2 = k, 3 = Absorption
            absorption = adachi_alpha.create_adachi_alpha(InLineComposition(layer), T=T, wl=wavelengths)[3]
        except:
            print("Warning: Using experimental data to estimate the absorption coefficient of material: ",
                  InLineComposition(layer))
            absorption = ToSolcoreMaterial(layer['properties']['composition'], T, execute=True).alpha(wavelengths)
            if layer['properties']['composition']['material'] == 'InGaAs':
                print(
                    "Warning: Extrapolation of experimental absorption data for InGaAs is not reliable at longer wavelengths.")
                print("    >>>: We truncate the absorption at the bandgap wavelength.")
                edge = 1240e-9 / (layer['properties']['band_gap'] / q)
                edgeidx = np.abs(wavelengths - edge).argmin()
                absorption[edgeidx:] = 0

    else:
        print(layer['properties']['composition'])
        absorption = ToSolcoreMaterial(layer['properties']['composition'], T, execute=True).alpha(wavelengths)
        try:

            if layer['properties']['composition']['material'] == 'InGaAs':
                print(
                    "Warning: Extrapolation of experimental absorption data for InGaAs is not reliable at longer wavelengths.")
                print("    >>>: We truncate the absorption at the bulk bandgap wavelength.")
                edge = 1240e-9 / (layer['properties']['band_gap'] / q)
                edgeidx = np.abs(wavelengths - edge).argmin()
                absorption[edgeidx:] = 0
        except Exception as err:
            print("Warning: Using Adachi calculation to estimate the absorption coefficient of material: ",
                  InLineComposition(layer))
            # 0 = Energy, 1 = n, 2 = k, 3 = Absorption
            try:
                absorption = adachi_alpha.create_adachi_alpha(InLineComposition(layer), T=T, wl=wavelengths)[3]
            except:
                print("Warning: No absorption information found for material {}. Setting it equal to zero.".format(
                    InLineComposition(layer)))
                absorption = 0 * wavelengths

    return [wavelengths.tolist(), absorption.tolist()]


def SolveQWproperties(device, calculate_absorption=True, WLsteps=(300e-9, 1100e-9, 201), wavelengths=None,
                      periodic=True, filter_strength=0.0, blur=None, blurmode="left", mode='kp8x8_bulk',
                      use_Adachi=False,
                      alpha_params=None):
    """ Considers the device as a QW and solves its properties, including the modification of the bandeges due to strain, the efective mases and the absorption coefficient. Without calling this function, the structure is just a colection of layers with bulk-like properties.

    :param device: The device structure
    :param calculate_absorption: If absorption must be calculated
    :param WLsteps: wavelengths in which to calculate the absorption (input for np.linspace function)
    :param wavelengths: An array with the waveengths
    :param periodic: If it has to be assumed that the structure is perdiodic
    :param filter_strength:
    :param blur:
    :param blurmode:
    :param mode:
    :param use_Adachi:
    :param alpha_params:
    :return: A dictionary with the output of the Schrodinger solver.
    """
    print('Solving QW properties...')
    T = device['T']

    QW = QWunit(ToStructure(device), substrate=ToSolcoreMaterial(device['substrate'], device['T'], execute=True))
    output = QW.solve(calculate_absorption=calculate_absorption, WLsteps=WLsteps, wavelengths=wavelengths,
                      T=device['T'], periodic=periodic, filter_strength=filter_strength, blur=blur, blurmode=blurmode,
                      mode=mode, use_Adachi=use_Adachi, alpha_params=alpha_params)

    for i in range(len(QW)):
        device['layers'][i]['properties']['band_gap'] = QW[i].eff_band_gap
        device['layers'][i]['properties']['electron_affinity'] = QW[i].eff_electron_affinity
        device['layers'][i]['properties']['eff_mass_electron_Gamma'] = QW[i].material.eff_mass_electron_Gamma
        device['layers'][i]['properties']['eff_mass_hh_z'] = QW[i].material.eff_mass_hh_z
        device['layers'][i]['properties']['eff_mass_lh_z'] = QW[i].material.eff_mass_lh_z
        device['layers'][i]['properties']['Nc'] = QW[i].material.Nc
        device['layers'][i]['properties']['Nv'] = QW[i].material.Nv
        device['layers'][i]['properties']['ni'] = np.sqrt(
            QW[i].material.Nc * QW[i].material.Nv * np.exp(-QW[i].eff_band_gap / (kb * T)))

        if calculate_absorption:
            device['layers'][i]['properties']['absorption'] = [QW.wl.tolist(), QW[i].material.absorption.tolist()]

    # Finally, we re-build a list of layers with the effective properties
    N = device['repeat']
    new_QW = []

    for i in range(len(QW)):
        # First, we create a dictionary with all the updated parameters
        param = dict(device['layers'][i]['properties'])
        del param['absorption']
        del param['composition']
        del param['width']

        # We recover the composition and thickness
        mat = device['layers'][i]['properties']['composition']
        width = device['layers'][i]['properties']['width']

        # Create the material with the updated properties
        layer_mat = ToSolcoreMaterial(mat, T, execute=True, **param)

        # In the end, we convert the absorption coefficient in extinction coefficient
        kk = QW[i].material.absorption * QW.wl / 4 / np.pi
        layer_mat.k = interp1d(QW.wl, kk, bounds_error=False, fill_value=(0, 0))
        layer_mat.n = QW[i].material.n
        # layer_mat.alpha = interp1d(QW.wl, QW[i].material.absorption, bounds_error=False, fill_value=(0, 0))
        import matplotlib.pyplot as plt
        plt.semilogy(QW.wl*1e9, layer_mat.n(QW.wl), label=i)
        plt.semilogy(QW.wl*1e9, layer_mat.k(QW.wl))

        # And the radiative recombination parameter
        inter = lambda E: layer_mat.n(E) ** 2 * layer_mat.alphaE(E) * np.exp(-E / (kb * T)) * E ** 2
        upper = layer_mat.band_gap + 10 * kb * T
        Br = 1.0 / layer_mat.ni ** 2 * 2 * pi / (h ** 3 * c ** 2) * quad(inter, 0, upper)[0]
        layer_mat.radiative_recombination = Br

        # And add the layer to the list of layers
        new_QW.append(Layer(width, layer_mat))

    # As the QW might be actually a MQW, we repeat this as many times as needed
    new_QW = N * new_QW

    plt.ylim(1e-5, 10)
    plt.legend()
    plt.show()
    import sys
    sys.exit()

    return new_QW


def CalculateAbsorptionProfile(z, wl, absorption):
    out = np.array(wl)
    out = np.vstack((out, absorption(z)))

    return out



import numpy
from numpy import *
from datetime import datetime
from solcore import spectral_conversion_nm_ev
from solcore.science_tracker import science_reference
import os

this_dir = os.path.split(__file__)[0]


def equation_of_time(day_angle):
    value = 229.18 * (0.000075 + 0.001868 * cos(day_angle) - 0.032077 * sin(day_angle) -
                      0.014615 * cos(2 * day_angle) - 0.040849 * sin(2 * day_angle))
    return value


def loadUtilitySpectra():
    global am_zero_wavelength, am_zero_irradiance, waterspectra, ozonespectra, uniformgasspectra

    spctra = numpy.loadtxt(os.path.join(this_dir, "SPCTRAL_si_units.txt"), unpack=True)
    (am_zero_wavelength, am_zero_irradiance, waterspectra, ozonespectra, uniformgasspectra) \
        = spctra
    waterspectra = waterspectra / 100
    ozonespectra = ozonespectra / 100


loadUtilitySpectra()


def get_default_spectral2_object():
    """ Returns a default state object for use with calculate_spectrum. Parameters returned are for:
        - Coventry (52.39°N,-1.56°E)
        - for 12:30 pm on the 30th of June 2011 (that's in the future!) (not any more..)
        - rural atmospheric conditions
        - 30% humidity
        - 1.42mm precipitated water
        - 1030 hPa atmospheric pressure
        - 0.34mm of Ozone
        - turbidity of 15%.
    """
    defaults = {}

    defaults["latitude"] = 52.39 / 180 * numpy.pi  # Coventry
    defaults["longitude"] = -1.56 / 180 * numpy.pi
    defaults["dateAndTime"] = datetime(2011, 6, 30, 12, 30)  # 12:30pm, 30th of June 2011
    defaults["aod_model"] = "rural"
    defaults["pressure"] = 103000.0  # si("1030 hPa")  #si ("1 atm")
    defaults["humidity"] = 0.3  # si("30 %")
    defaults["precipwater"] = 0.00142  # si("1.42 mm")    # unit?
    defaults["ozone"] = 0.00034  # si("0.34 mm")          # unit?
    defaults["turbidity"] = 0.15  # unit?

    defaults["display units"] = {
        "pressure": ("hPa", 0),
        "latitude": (u"degrees", 3),
        "longitude": (u"degrees", 3),
        "precipwater": ("cm", 3),
        "ozone": ("cm", 3),
        "humidity": ("%", 1),
    }

    return defaults


def calculate_spectrum_spectral2(stateObject=None, suppress_nan=True, power_density_in_nm=False):
    """ Calculates a solar spectrum using the SPECTRAL2 irradiance model developed by the NRL:

    "http://rredc.nrel.gov/solar/models/spectral/SPCTRAL2/"
    
    :param stateObject:
    :param suppress_nan:
    :return:
    """
    science_reference("spectral2 irradiance model", "http://rredc.nrel.gov/solar/models/spectral/SPCTRAL2/")

    if stateObject == None:
        stateObject = get_default_spectral2_object()

    latitude = stateObject["latitude"]
    longitude = stateObject["longitude"]
    dateAndTime = stateObject["dateAndTime"]
    aod_model = stateObject["aod_model"]
    pressure = stateObject["pressure"]
    humidity = stateObject["humidity"]
    ozone = stateObject["ozone"]
    turbidity = stateObject["turbidity"]
    precipwater = stateObject["precipwater"]

    assert aod_model in "rural urban maritime tropospheric".split(), "aod_model must be rural, urban, maritime, or tropospheric"

    time_since_new_years = dateAndTime - datetime(dateAndTime.year, 1, 1, 0, 0)  # 1/1/year, 0:00 am.
    time_since_midnight = dateAndTime - datetime(dateAndTime.year, dateAndTime.month, dateAndTime.day, 0, 0)
    # hours_since_midnight = time_since_midnight.seconds / convert(time_since_midnight.seconds, "s", "h")
    hours_since_midnight = time_since_midnight.seconds / 3600
    day_number = time_since_new_years.days

    # converting back go degrees for longitude rounding
    longitude_degrees = longitude / numpy.pi * 180  # convert(longitude,"radians",u"degrees")

    day_angle = (2.0 * pi * (day_number - 1.0)) / 365.0  # this is in radians
    hour_angle_degrees = 15.0 * (hours_since_midnight + equation_of_time(day_angle) / 60.0 +
                                 ((int(longitude_degrees / 15.0) * 15.0 - longitude_degrees) * 4.0)
                                 / 60.0 + 12.0) - 360.0  # this is in degrees.

    hour_angle = hour_angle_degrees / 180 * numpy.pi  # convert(hour_angle_degrees, "degrees", "radians")

    declination = (
        0.006918
        - 0.399912 * cos(day_angle)
        + 0.070257 * sin(day_angle)
        - 0.006758 * cos(2 * day_angle)
        + 0.000907 * sin(2 * day_angle)
        - 0.002697 * cos(3 * day_angle)
        + 0.00148 * sin(3 * day_angle)
    )

    earth_sun_distance_factor = (
        1.00011
        + 0.034221 * cos(day_angle)
        + 0.001280 * sin(day_angle)
        + 0.000719 * cos(2 * day_angle)
        + 0.000077 * sin(2 * day_angle)
    )

    solar_zenith_angle = arccos(
        cos(declination) * cos(latitude) * cos(hour_angle)
        + sin(declination) * sin(latitude)
    )

    solar_zenith_angle_degrees = solar_zenith_angle / pi * 180  # convert(solar_zenith_angle,"radians",u'degrees')
    # original code checked to stop sun dropping below horizon
    # //Stop sun dropping below horizon
    #     if(solar_zenith_angle > 91)
    #        solar_zenith_angle = 91; (was in degrees)


    # this used to be 1/ could this cause issues in java?
    relative_am = 1.0 / (cos(solar_zenith_angle) + (
        0.15 * pow((93.885 - solar_zenith_angle_degrees), -1.253)))  ##AARG raised to the power of degrees
    pressure_corrected_am = relative_am * (pressure / 101325.33538686013)  # si("1 atm")
    effective_ozone_am = (1.0 + (22.0 / 6370.0)) / (
        pow((pow(cos(solar_zenith_angle), 2.0) + (2.0 * (22.0 / 6370.0))), 0.5))

    if aod_model == "rural":
        c_coefficient = [0.581, 16.823, 17.539]
        d_coefficient = [0.8547, 78.696, 0, 64.458]
    elif aod_model == "rural":
        c_coefficient = [0.2595, 33.843, 39.524]
        d_coefficient = [1.0, 84.254, -9.1, 65.458]
    elif aod_model == "maritime":
        c_coefficient = [0.1134, 0.8941, 1.0796]
        d_coefficient = [0.04435, 1.6048, 0, 1.5298];
    elif aod_model == "tropospheric":
        c_coefficient = [0.6786, 13.899, 13.313]
        d_coefficient = [1.8379, 14.912, 0, 5.96]
    elif type(aod_model) == list:
        c_coefficient, d_coefficient = aod_model

    # 0.9 degrees * something_in_percent rather than 90 degrees?? Honestly.
    x_humidity = cos(humidity * pi / 2.)  # si(0.9,"degrees","radians")

    alpha1 = (c_coefficient[0] + c_coefficient[1] * x_humidity) / (1 + c_coefficient[2] * x_humidity)
    alpha2 = (d_coefficient[0] + d_coefficient[1] * x_humidity + d_coefficient[2] * x_humidity ** 2) / (
        1 + (d_coefficient[3] * x_humidity))

    wavelength_um = am_zero_wavelength * 1e6  # convert(am_zero_wavelength, "m", 'um')

    rayleigh_coeff = exp(-pressure_corrected_am / (wavelength_um ** 4 * (115.6406 - 1.335 / wavelength_um ** 2)))

    aerosol_coeff = where(wavelength_um <= 0.5,
                          exp(-pressure_corrected_am * turbidity * 2.0 ** (alpha2 - alpha1) * wavelength_um ** -alpha1),
                          exp(-pressure_corrected_am * turbidity * wavelength_um ** -alpha2)
                          )

    vapour_coeff = exp(-(0.2385 * precipwater * waterspectra * relative_am)
                       / (1. + 20.07 * precipwater * waterspectra * relative_am) ** 0.45)

    ozone_coeff = exp(-ozonespectra * ozone * effective_ozone_am)

    mixed_coeff = exp((-1.41 * uniformgasspectra * pressure_corrected_am)
                      / (1. + 118.93 * uniformgasspectra * pressure_corrected_am) ** 0.45);

    # Apply attenuation/scatting coefficients to the per m specturm (SI wavelength spectrum)
    wavelength_m = am_zero_wavelength
    irradiance_per_m = am_zero_irradiance * earth_sun_distance_factor * rayleigh_coeff * aerosol_coeff * vapour_coeff * ozone_coeff * mixed_coeff

    if suppress_nan:
        irradiance_per_m = numpy.nan_to_num(irradiance_per_m)

    storage = dict()

    # Convert to per nm spectrum
    wavelength_nm = wavelength_m * 1e9  # asUnit(wavelength_m,"nm")
    irradiance_per_nm = 1e-9 * irradiance_per_m  # asUnit(irradiance_per_m,"nm-1")

    if power_density_in_nm:
        storage = wavelength_nm, irradiance_per_nm

    else:
        # Convert to per eV spectrum
        energy_ev, irradiance_per_ev = spectral_conversion_nm_ev(wavelength_nm, irradiance_per_nm)

        # Covert to energy per Joule spectrum
        energy_j = energy_ev * 1.6e-19  # siUnits(energy_ev, 'J')
        irradiance_per_j = irradiance_per_ev / 1.6e-19  # siUnits(irradiance_per_ev, 'J-1')

        # integrate to find power density
        power_density = trapz(x=wavelength_m, y=irradiance_per_m)

        storage['incident power density'] = power_density
        storage['incident spectrum wavelength si'] = array([wavelength_m, irradiance_per_m])
        storage['incident spectrum wavelength nm'] = array([wavelength_nm, irradiance_per_nm])
        storage['incident spectrum energy si'] = array([energy_j, irradiance_per_j])
        storage['incident spectrum energy eV'] = array([energy_ev, irradiance_per_ev])

    return storage


if __name__ == "__main__":
    data = calculate_spectrum_spectral2()

    import matplotlib.pyplot as plt

    plt.plot(data['incident spectrum wavelength nm'][0], data['incident spectrum wavelength nm'][1])
    plt.show()



import numpy as np
from solcore.state import State


def test_state():
    array1 = np.array([1, 2, 3])
    array2 = np.array([2, 3, 4])

    state1 = State()
    state1.a = 1
    state1.b = True
    state1.c = array1
    assert state1.__len__() == 3
    assert state1.a == 1
    assert state1.b == True
    assert all([input == output for input, output in zip(array1, state1.c)])

    state2 = State()
    state2.b = False
    state2.d = array2
    state1.update(state2)
    assert state1.__len__() == 4
    assert state1.a == 1
    assert state2.b == False
    assert all([input == output for input, output in zip(array1, state1.c)])
    assert all([input == output for input, output in zip(array2, state1.d)])



from .beer_lambert import solve_beer_lambert
from .tmm import solve_tmm
from .external_optics import solve_external_optics

try:
    from .rcwa import solve_rcwa, rcwa_options
except ImportError:
    rcwa_options = {}
    solve_rcwa = None


def VBO_align(structure):
    """Takes in the the structure layers and sets Ec and Ev energies in the state which aligns the heterostructure."""

    for layer in structure:
        layer.Ev = layer.material.valence_band_offset
        layer.Ec = layer.material.valence_band_offset + layer.material.band_gap
    return structure



import ast  # completely safe eval, but can't call functions
import itertools  # iterate over element/fractions in material string parser
import math  # hyperbolic functions etc in parameterisation
import re  # parsing material string
from functools import lru_cache  # cache function calls to stop things taking forever / recalculating smae things
import numpy
from typing import Optional, Callable

import solcore
from solcore import siUnits
from solcore.source_managed_class import SourceManagedClass


def safe_cacher(maxsize):
    def safewrap(uncached):
        cached = lru_cache(maxsize=maxsize)(uncached)

        def mucked_up_func(*arg, **kwarg):
            try:
                return cached(*arg, **kwarg)
            except:
                return uncached(*arg, **kwarg)

        mucked_up_func.cache_clear = cached.cache_clear
        return mucked_up_func

    return safewrap


def grouper(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)  # python2: izip_longest


def bow(parent_0_value, parent_1_value, bowing_parameter, x):
    return parent_0_value * (1 - x) + parent_1_value * x - bowing_parameter * (1 - x) * x


class ParameterSystem(SourceManagedClass):
    """Parameter database/bowing system for compound materials.
    
    Once instantiated, this plugin loads the materials parameterisations Parameters for
    compound materials can be retrieved with the get_parameter function.
    """

    def __init__(self, sources: Optional[Callable] = None):
        super().__init__({k: sources(k) for k in sources()})
        # create a dictionary that's safe for the eval function to use, so that config
        # files don't have access to all of python
        self.__assemble_builtins()
        # Matches capital letter + n * small letter, e.g.: In, Ga, As
        self.element_RE = re.compile(
            "([A-Z][a-z]*)")

    def get_parameter(self, material, parameter, verbose=False, **others):
        """Calculate/look up parameters for materials, returns in SI units
        
        Usage: .get_parameter(material_name, parameter_name, **kwargs)
        - material_name is a string of element symbols/fractions, e.g.: In0.2GaAsP0.1
        - parameter_name is a string of 
        - **kwargs captures parameters that may be necessary for some calculations, eg. Temperature
            material fractions may also be specified here, e.g.: .get_parameter("InGaAs", "band_gap", In=0.2)
        
        If a compound material is bowed between two parent materials, the parent materials' parameters are calculated
        recursively with this function. The final parameter is calculated as:
            result=parent_0_value * (1-x) + parent_1_value*x - bowing_parameter * (1-x) * x
        
        The function is cached, so that multiple calls with the same parameters do not incur additional overhead.
        
        """
        material, relevant_parameters = self.__parse_material_string(material, others)

        def tryget(p, alternative):
            try:
                result = self.get_parameter(material, p, verbose=verbose, **relevant_parameters)
                return result
            except ValueError:
                return alternative

        calculation_environment = {
            "get": lambda p: self.get_parameter(material, p, verbose=verbose, **relevant_parameters),
            "tget": tryget

        }

        if verbose:
            print(material, parameter, relevant_parameters)

        assert material in self.database.sections(), "Material {} not in database".format(material)

        if "Final Calculables" in self.database.sections() and \
                        parameter in self.database.options(
                    "Final Calculables") and not parameter in self.database.options(
            material):  # since this is recursive, "Final" gets done first!
            expression = self.database.get("Final Calculables", parameter)
            calculation_environment.update(relevant_parameters)
            result = self.__eval_string_expression(expression, **calculation_environment)
            return result

        if "x" in self.database.options(material):  # material is bowed
            bowed_element = self.database.get(material, "x")
            x = relevant_parameters[bowed_element]
            # del relevant_parameters[bowed_element] # not propagating the element reduces cache misses
            parent0 = self.database.get(material, "parent0")
            parent1 = self.database.get(material, "parent1")

            if parameter in self.database.options(material):
                bowing_parameter = self.__eval_string_expression(
                    self.database.get(material, parameter),
                    **relevant_parameters)
            else:
                bowing_parameter = 0
            del relevant_parameters[bowed_element]  # not propagating the element reduces cache misses

            parent0_value = self.get_parameter(parent0, parameter, verbose=verbose, **relevant_parameters)
            parent1_value = self.get_parameter(parent1, parameter, verbose=verbose, **relevant_parameters)
            return bow(parent0_value, parent1_value, bowing_parameter, x)

        if parameter in self.database.options(material):
            return self.__eval_string_expression(self.database.get(material, parameter), **relevant_parameters)

        if "Immediate Calculables" in self.database.sections() and \
                        parameter in self.database.options("Immediate Calculables"):
            expression = self.database.get("Immediate Calculables", parameter)
            calculation_environment.update(relevant_parameters)
            result = self.__eval_string_expression(expression, **calculation_environment)
            return result

        raise ValueError(
            "Parameter '{}' not in material '{}', nor in calculable parameters.".format(parameter, material))

    def __parse_material_string(self, material_string, other_parameters):
        """parses the material identifier strings of these types:
        
            - In0.2GaAsP0.01
            - InGaAsP {"In":0.2, "P":0.01}
            
            into:
                tuple("InGaAsP", {"In":0.2, "P":0.01})
            
            other parameters are passed into the fractions dictionary. Chemical element Symbols are permitted as 
            sub-material strings, as well as longer words as long as they begin with a capital letter. 
        """
        if "{" in material_string:  # fractions given as a dictionary: InGaAsP {'In':0.2, 'P':0.01}
            identifier, arguments = material_string.split(" ")
            arguments = ast.literal_eval(arguments)
            assert type(arguments) == dict, "{} is not a dict".format(arguments)
            arguments.update(other_parameters)
            return identifier, arguments

        # else: fractions given in parameter or in string: In0.2GaAsP0.01
        elements_and_fractions = self.element_RE.split(material_string)[1:]
        arguments = {}
        for element, fraction in grouper(elements_and_fractions, 2):
            try:
                arguments[element] = float(fraction)
            except:
                pass
        arguments.update(other_parameters)
        # print ("".join(elements_and_fractions[::2]),arguments )
        return "".join(elements_and_fractions[::2]), arguments

    def __eval_string_expression(self, string_expression, **others):
        if " " in string_expression:  # treat second part as unit!
            string_expression, units = string_expression.split(" ", 1)
            use_units = True
        else:
            use_units = False

        if 'T' in string_expression and 'T' not in others.keys():
            raise KeyError('The temperature is needed to calculate this parameter. '
                             'Include keyword argument "T" when calling "get_parameter"')

        non_converted_unit = eval(string_expression, {"__builtins__": self.builtins_replacement}, others)
        in_si_units = siUnits(non_converted_unit, units) if use_units else non_converted_unit
        return in_si_units

    def __assemble_builtins(self):
        self.builtins_replacement = {"max": numpy.max, "min": numpy.min}
        self.builtins_replacement.update(math.__dict__)


if __name__ == "__main__":
    import os

    v = ParameterSystem()
    v.add_source("v", os.path.split(__file__)[0] + "/plugins/vurgaftman/builtins/endpoints.txt")
    v.add_source("v2", os.path.split(__file__)[0] + "/plugins/vurgaftman/builtins/bowing_tree.txt")
    print(solcore.asUnit(v.get_parameter("GaAsSb.5", "band_gap", verbose=True, T=300), "eV"))
    print((v.get_parameter("GaAsSb.75", "band_gap", verbose=True, T=300), "eV"))
    print((v.get_parameter("GaAsSb.5", "band_gap", verbose=True, T=300), "eV"))



# Package file for the Poisson - Drift-Diffusion solver

# We import all the contents of the modules of this package so they can be used directly
from .DeviceStructure import *
from .QWunit import *

try:
    from .DriftDiffusionUtilities import *
except Exception as err:
    print('WARNING: The Poisson - Drift-Diffusion solver will not be available because '
          'the ddModel fortran library could not be imported.')
    print(err)




from pytest import approx, mark, raises
import numpy as np
from solcore.constants import kb, q, vacuum_permittivity



def test_get_j_dark():
    from solcore.analytic_solar_cells.depletion_approximation import get_j_dark

    x = np.power(10, np.random.uniform(-8, -5))
    xi = np.power(10, np.random.uniform(-8, -5))

    l = np.power(10, np.random.uniform(-9, -6))
    s = np.power(10, np.random.uniform(1, 3))
    d = np.power(10, np.random.uniform(-5, 0))
    Vbi = np.random.uniform(0.1,5)
    minor = np.power(10, np.random.uniform(-7, -4))
    T = np.random.uniform(0.1,400)
    es = np.random.uniform(1,20)*vacuum_permittivity
    Na = np.power(10, np.random.uniform(22, 25))
    Nd = np.power(10, np.random.uniform(22, 25))

    V = np.linspace(-6, 4, 20)
    V = np.where(V < Vbi - 0.001, V, Vbi - 0.001)

    wn = (-xi + np.sqrt(xi ** 2 + 2. * es * (Vbi - V) / q * (1 / Na + 1 / Nd))) / (1 + Nd / Na)
    wp = (-xi + np.sqrt(xi ** 2 + 2. * es * (Vbi - V) / q * (1 / Na + 1 / Nd))) / (1 + Na / Nd)

    w = wn + wp + xi

    expected = (q*d*minor/l)*(np.exp(q*V/(kb*T))-1)*((((s*l)/d)*np.cosh((x-w)/l)+np.sinh((x-w)/l))/
                                                  (((s*l)/d)*np.sinh((x-w)/l)+np.cosh((x-w)/l)))

    result = get_j_dark(x, w, l, s, d, V, minor, T)

    assert result == approx(expected, nan_ok=True)


def test_factor():
    from solcore.analytic_solar_cells.depletion_approximation import factor
    T = np.random.uniform(0.1,400)
    Vbi = np.random.uniform(0.1,5)
    tp = np.power(10, np.random.uniform(-10, -5))
    tn = np.power(10, np.random.uniform(-10, -5))
    dEt = 0

    kT = kb*T
    V = np.linspace(-6, 4, 20)
    V = np.where(V < Vbi - 0.001, V, Vbi - 0.001)

    m = V >= -1120 * kT / q
    V = V[m]

    b= np.exp(-V*q/(2*kT))*np.cosh((q*dEt/kT)+(1/2)*np.log(tp/tn))

    z1 = np.sqrt(tp/tn)*np.exp(-q*(Vbi - V)/(kT * 2))
    z2 = np.sqrt(tp/tn)*np.exp(q*(Vbi - V)/(kT * 2))

    expected = np.zeros(b.shape)

    # For b values < 1
    inds = b < 1  # use tan-1 formulation without issue

    upper = np.arctan((b[inds]+z2[inds])/np.sqrt(1-b[inds]**2))
    lower = np.arctan((b[inds]+z1[inds])/np.sqrt(1-b[inds]**2))

    expected[inds] = (upper - lower)/np.sqrt(1-b[inds]**2)

    # for b values >=1, log formulation
    inds = (b >= 1) & (b <= 1e7)

    upper = np.log(abs(z2[inds] + b[inds] - np.sqrt(b[inds]**2-1))) - \
            np.log(abs(z2[inds] + b[inds] + np.sqrt(b[inds]**2-1)))

    lower = np.log(abs(z1[inds] + b[inds] - np.sqrt(b[inds]**2-1))) - \
            np.log(abs(z1[inds] + b[inds] + np.sqrt(b[inds]**2-1)))
    expected[inds] = (upper - lower)/(2*np.sqrt(b[inds]**2 - 1))

    # limit as b gets very large: sqrt(b**2-1) = b

    inds = b > 1e7

    upper = np.log(z2[inds]) - np.log(z2[inds] + 2*b[inds])
    lower = np.log(z1[inds]) - np.log(z1[inds] + 2*b[inds])

    expected[inds] = (upper - lower) / (2*b[inds])

    result = factor(V, Vbi, tp, tn, kT, dEt)

    assert result == approx(expected, nan_ok=True)


def test_forward():
    from solcore.analytic_solar_cells.depletion_approximation import factor, forward

    T = np.random.uniform(0.1,400)
    Vbi = np.random.uniform(0.1,5)
    tp = np.power(10, np.random.uniform(-10, -5))
    tn = np.power(10, np.random.uniform(-10, -5))
    dEt = 0

    kT = kb*T
    V = np.linspace(-6, 4, 20)
    V = np.where(V < Vbi - 0.001, V, Vbi - 0.001)

    ni = np.power(10, np.random.uniform(2, 9))
    es = np.random.uniform(1,20)*vacuum_permittivity
    Na = np.power(10, np.random.uniform(22, 25))
    Nd = np.power(10, np.random.uniform(22, 25))
    xi = np.power(10, np.random.uniform(-8, -5))

    m = V >= -1120 * kT / q
    V = V[m]

    wn = (-xi + np.sqrt(xi ** 2 + 2. * es * (Vbi - V) / q * (1 / Na + 1 / Nd))) / (1 + Nd / Na)
    wp = (-xi + np.sqrt(xi ** 2 + 2. * es * (Vbi - V) / q * (1 / Na + 1 / Nd))) / (1 + Na / Nd)

    w = wn + wp + xi

    f_b = factor(V, Vbi, tp, tn, kT, dEt)
    expected = 2 * q * ni * w / np.sqrt(tn * tp) * \
          np.sinh(q*V / (2 * kT)) / (q * (Vbi - V) / kT) * f_b

    result = forward(ni, V, Vbi, tp, tn, w, kT, dEt)

    assert result == approx(expected, nan_ok=True)


def test_get_J_srh():
    from solcore.analytic_solar_cells.depletion_approximation import forward, get_Jsrh

    T = np.random.uniform(0.1,400)
    Vbi = np.random.uniform(0.1,5)
    tp = np.power(10, np.random.uniform(-10, -5))
    tn = np.power(10, np.random.uniform(-10, -5))
    dEt = 0

    kT = kb*T
    V = np.linspace(-6, 4, 20)
    V = np.where(V < Vbi - 0.001, V, Vbi - 0.001)

    ni = np.power(10, np.random.uniform(2, 9))
    es = np.random.uniform(1,20)*vacuum_permittivity
    Na = np.power(10, np.random.uniform(22, 25))
    Nd = np.power(10, np.random.uniform(22, 25))
    xi = np.power(10, np.random.uniform(-8, -5))

    wn = (-xi + np.sqrt(xi ** 2 + 2. * es * (Vbi - V) / q * (1 / Na + 1 / Nd))) / (1 + Nd / Na)
    wp = (-xi + np.sqrt(xi ** 2 + 2. * es * (Vbi - V) / q * (1 / Na + 1 / Nd))) / (1 + Na / Nd)

    w = wn + wp + xi

    expected = np.zeros(V.shape)

    inds = V >= -1120 * kT / q
    expected[inds] = forward(ni, V[inds], Vbi, tp, tn, w[inds], kT)

    result = get_Jsrh(ni, V, Vbi, tp, tn, w, kT, dEt)

    assert result == approx(expected, nan_ok=True)


def test_get_J_sc_diffusion_top():
    from solcore.analytic_solar_cells.depletion_approximation import get_J_sc_diffusion
    from scipy.integrate import solve_bvp
    from solcore.interpolate import interp1d
    from solcore.light_source import LightSource

    D = np.power(10, np.random.uniform(-5, 0)) # Diffusion coefficient
    L = np.power(10, np.random.uniform(-9, -6)) # Diffusion length
    minority = np.power(10, np.random.uniform(-7, -4))# minority carrier density
    s = np.power(10, np.random.uniform(0, 3)) # surface recombination velocity

    light_source = LightSource(source_type="standard", version="AM1.5g")
    wl = np.linspace(300, 1800, 50)*1e-9
    wl_ls, phg = light_source.spectrum(output_units='photon_flux_per_m', x=wl)

    xa_nm = np.random.uniform(1, 1000)
    xa = xa_nm*1e-9
    xb = np.random.uniform(xa_nm+1, 1100)*1e-9

    ## make a simple Beer-Lambert profile
    dist = np.linspace(0, xb, 1000)
    alphas = np.linspace(1e8, 10, len(wl))

    expn = np.exp(- alphas[:, None] * dist[None,:])

    output = alphas[:, None]*expn
    output = output.T
    gen_prof = interp1d(dist, output, axis = 0)


    zz = np.linspace(xa, xb, 1002)[:-1]
    gg = gen_prof(zz) * phg

    g_vs_z = np.trapz(gg, wl, axis=1)

    A = lambda x: np.interp(x, zz, g_vs_z) / D + minority / L ** 2

    def fun(x, y):
        out1 = y[1]
        out2 = y[0] / L ** 2 - A(x)
        return np.vstack((out1, out2))

    def bc(ya, yb):
        left = ya[1] - s / D * (ya[0] - minority)
        right = yb[0]
        return np.array([left, right])


    guess = minority * np.ones((2, zz.size))
    guess[1] = np.zeros_like(guess[0])

    solution = solve_bvp(fun, bc, zz, guess)

    expected = solution.y[1][-1]

    result = get_J_sc_diffusion(xa, xb, gen_prof, D, L, minority, s, wl, phg, side='top')

    assert result == approx(expected)


def test_get_J_sc_diffusion_bottom():
    from solcore.analytic_solar_cells.depletion_approximation import get_J_sc_diffusion
    from scipy.integrate import solve_bvp
    from solcore.interpolate import interp1d
    from solcore.light_source import LightSource

    D = np.power(10, np.random.uniform(-5, 0)) # Diffusion coefficient
    L = np.power(10, np.random.uniform(-9, -6)) # Diffusion length
    minority = np.power(10, np.random.uniform(-7, -4))# minority carrier density
    s = np.power(10, np.random.uniform(0, 3)) # surface recombination velocity

    light_source = LightSource(source_type="standard", version="AM1.5g")
    wl = np.linspace(300, 1800, 50)*1e-9
    wl_ls, phg = light_source.spectrum(output_units='photon_flux_per_m', x=wl)

    xa_nm = np.random.uniform(1, 1000)
    xa = xa_nm*1e-9
    xb = np.random.uniform(xa_nm+1, 1100)*1e-9

    ## make a simple Beer-Lambert profile
    dist = np.linspace(0, xb, 1000)
    alphas = np.linspace(1e8, 10, len(wl))

    expn = np.exp(- alphas[:, None] * dist[None,:])

    output = alphas[:, None]*expn
    output = output.T
    gen_prof = interp1d(dist, output, axis = 0)


    zz = np.linspace(xa, xb, 1002)[:-1]
    gg = gen_prof(zz) * phg

    g_vs_z = np.trapz(gg, wl, axis=1)

    A = lambda x: np.interp(x, zz, g_vs_z) / D + minority / L ** 2

    def fun(x, y):
        out1 = y[1]
        out2 = y[0] / L ** 2 - A(x)
        return np.vstack((out1, out2))

    def bc(ya, yb):
        left = ya[0]
        right = yb[1] - s / D * (yb[0] - minority)
        return np.array([left, right])


    guess = minority * np.ones((2, zz.size))
    guess[1] = np.zeros_like(guess[0])

    solution = solve_bvp(fun, bc, zz, guess)

    expected = solution.y[1][0]

    result = get_J_sc_diffusion(xa, xb, gen_prof, D, L, minority, s, wl, phg, side='bottom')

    assert result == approx(expected)


def test_get_J_sc_SCR():
    from solcore.light_source import LightSource
    from solcore.interpolate import interp1d
    from solcore.analytic_solar_cells.depletion_approximation import get_J_sc_SCR

    light_source = LightSource(source_type="standard", version="AM1.5g")
    wl = np.linspace(300, 1800, 50)*1e-9
    wl_ls, phg = light_source.spectrum(output_units='photon_flux_per_m', x=wl)

    xa_nm = np.random.uniform(1, 1000)
    xa = xa_nm*1e-9
    xb = np.random.uniform(xa_nm+1, 1100)*1e-9

    ## make a simple Beer-Lambert profile
    dist = np.linspace(0, xb, 1000)
    alphas = np.linspace(1e5, 10, len(wl))

    expn = np.exp(- alphas[:, None] * dist[None,:])

    output = alphas[:, None]*expn
    output = output.T
    gen_prof = interp1d(dist, output, axis = 0)

    zz = np.linspace(xa, xb, 1002)[:-1]
    gg = gen_prof(zz) * phg
    expected = np.trapz(np.trapz(gg, wl, axis=1), zz)

    result = get_J_sc_SCR(xa, xb, gen_prof, wl, phg)

    assert  expected == approx(result)


def test_get_J_sc_SCR_vs_WL():
    from solcore.light_source import LightSource
    from solcore.interpolate import interp1d
    from solcore.analytic_solar_cells.depletion_approximation import get_J_sc_SCR_vs_WL

    light_source = LightSource(source_type="standard", version="AM1.5g")
    wl = np.linspace(300, 1800, 50)*1e-9
    wl_ls, phg = light_source.spectrum(output_units='photon_flux_per_m', x=wl)

    xa_nm = np.random.uniform(1, 1000)
    xa = xa_nm*1e-9
    xb = np.random.uniform(xa_nm+1, 1100)*1e-9

    ## make a simple Beer-Lambert profile
    dist = np.linspace(0, xb, 1000)
    alphas = np.linspace(1e5, 10, len(wl))

    expn = np.exp(- alphas[:, None] * dist[None,:])

    output = alphas[:, None]*expn
    output = output.T
    gen_prof = interp1d(dist, output, axis = 0)

    zz = np.linspace(xa, xb, 1002)[:-1]
    gg = gen_prof(zz) * phg
    expected = np.trapz(gg, zz, axis=0)

    result = get_J_sc_SCR_vs_WL(xa, xb, gen_prof, wl, phg)

    assert  expected == approx(result)


def test_get_J_sc_diffusion_vs_WL_top():
    from solcore.analytic_solar_cells.depletion_approximation import get_J_sc_diffusion_vs_WL
    from scipy.integrate import solve_bvp
    from solcore.interpolate import interp1d
    from solcore.light_source import LightSource

    D = np.power(10, np.random.uniform(-5, 0)) # Diffusion coefficient
    L = np.power(10, np.random.uniform(-9, -6)) # Diffusion length
    minority = np.power(10, np.random.uniform(-7, -4))# minority carrier density
    s = np.power(10, np.random.uniform(0, 3)) # surface recombination velocity

    light_source = LightSource(source_type="standard", version="AM1.5g")
    wl = np.linspace(300, 1800, 50)*1e-9
    wl_ls, phg = light_source.spectrum(output_units='photon_flux_per_m', x=wl)

    xa_nm = np.random.uniform(1, 1000)
    xa = xa_nm*1e-9
    xb = np.random.uniform(xa_nm+1, 1100)*1e-9

    ## make a simple Beer-Lambert profile
    dist = np.linspace(0, xb, 1000)
    alphas = np.linspace(1e8, 10, len(wl))

    expn = np.exp(- alphas[:, None] * dist[None,:])

    output = alphas[:, None]*expn
    output = output.T
    gen_prof = interp1d(dist, output, axis = 0)


    zz =  np.linspace(xa, xb, 1002)[:-1]
    gg = gen_prof(zz) * phg

    expected = np.zeros_like(wl)

    for i in range(len(wl)):

        if np.all(gg[:,i] == 0): # no reason to solve anything if no generation at this wavelength
            expected[i] = 0

        A = lambda x: np.interp(x, zz, gg[:, i]) / D + minority / L ** 2

        def fun(x, y):
            out1 = y[1]
            out2 = y[0] / L ** 2 - A(x)
            return np.vstack((out1, out2))

        def bc(ya, yb):
            left = ya[1] - s / D * (ya[0] - minority)
            right = yb[0]
            return np.array([left, right])

        guess = minority * np.ones((2, zz.size))
        guess[1] = np.zeros_like(guess[0])
        solution = solve_bvp(fun, bc, zz, guess)

        expected[i] = solution.y[1][-1]

    result = get_J_sc_diffusion_vs_WL(xa, xb, gen_prof, D, L, minority, s, wl, phg, side='top')


    assert result == approx(expected)


def test_get_J_sc_diffusion_vs_WL_bottom():
    from solcore.analytic_solar_cells.depletion_approximation import get_J_sc_diffusion_vs_WL
    from scipy.integrate import solve_bvp
    from solcore.interpolate import interp1d
    from solcore.light_source import LightSource

    D = np.power(10, np.random.uniform(-5, 0)) # Diffusion coefficient
    L = np.power(10, np.random.uniform(-9, -6)) # Diffusion length
    minority = np.power(10, np.random.uniform(-7, -4))# minority carrier density
    s = np.power(10, np.random.uniform(0, 3)) # surface recombination velocity

    light_source = LightSource(source_type="standard", version="AM1.5g")
    wl = np.linspace(300, 1800, 50) * 1e-9
    wl_ls, phg = light_source.spectrum(output_units='photon_flux_per_m', x=wl)

    xa_nm = np.random.uniform(1, 1000)
    xa = xa_nm * 1e-9
    xb = np.random.uniform(xa_nm + 1, 1100) * 1e-9

    ## make a simple Beer-Lambert profile
    dist = np.linspace(0, xb, 1000)
    alphas = np.linspace(1e8, 10, len(wl))

    expn = np.exp(- alphas[:, None] * dist[None, :])

    output = alphas[:, None] * expn
    output = output.T
    gen_prof = interp1d(dist, output, axis=0)

    zz = np.linspace(xa, xb, 1002)[:-1]
    gg = gen_prof(zz) * phg

    expected = np.zeros_like(wl)

    for i in range(len(wl)):
        if np.all(gg[:, i] == 0):  # no reason to solve anything if no generation at this wavelength
            expected[i] = 0

        A = lambda x: np.interp(x, zz, gg[:, i]) / D + minority / L ** 2

        def fun(x, y):
            out1 = y[1]
            out2 = y[0] / L ** 2 - A(x)
            return np.vstack((out1, out2))

        def bc(ya, yb):
            left = ya[0]
            right = yb[1] - s / D * (yb[0] - minority)
            return np.array([left, right])

        guess = minority * np.ones((2, zz.size))
        guess[1] = np.zeros_like(guess[0])
        solution = solve_bvp(fun, bc, zz, guess)

        expected[i] = solution.y[1][0]

    result = get_J_sc_diffusion_vs_WL(xa, xb, gen_prof, D, L, minority, s, wl, phg, side='bottom')

    assert result == approx(expected)


def test_identify_layers_exceptions():
    from solcore.analytic_solar_cells.depletion_approximation import identify_layers
    from solcore import material
    from solcore.structure import Layer, Junction

    Na = np.power(10, np.random.uniform(22, 25))
    Nd = np.power(10, np.random.uniform(22, 25))

    Lp = np.power(10, np.random.uniform(-9, -6))# Diffusion length
    Ln = np.power(10, np.random.uniform(-9, -6)) # Diffusion length

    GaAs_n = material("GaAs")(Nd = Nd, hole_diffusion_length=Ln)
    GaAs_p = material("GaAs")(Na = Na, electron_diffusion_length=Lp)
    GaAs_i = material("GaAs")()
    Ge_n = material("Ge")(Nd = Nd, hole_diffusion_length=Ln)

    n_width = np.random.uniform(500, 1000)*1e-9
    p_width = np.random.uniform(3000, 5000)*1e-9
    i_width = np.random.uniform(300, 500) * 1e-9

    test_junc  = Junction([Layer(n_width, GaAs_n,role="emitter"),
                           Layer(p_width, GaAs_p, role="neither")])

    with raises(RuntimeError):
        identify_layers(test_junc)

    test_junc =  Junction([Layer(n_width, GaAs_n,role="emitter"),
                           Layer(i_width, GaAs_i, role="intrinsic"),
                           Layer(p_width, GaAs_p, role="nothing")])

    with raises(RuntimeError):
        identify_layers(test_junc)

    test_junc  = Junction([Layer(n_width, Ge_n,role="emitter"),
                           Layer(p_width, GaAs_p, role="base")])

    with raises(AssertionError):
        identify_layers(test_junc)


def test_process_junction_np():
    from solcore.analytic_solar_cells.depletion_approximation import identify_layers, identify_parameters
    from solcore import material
    from solcore.structure import Layer, Junction
    from solcore.state import State

    Na = np.power(10, np.random.uniform(23, 26))
    Nd = np.power(10, np.random.uniform(22, 25))

    options = State()
    options.T = np.random.uniform(250, 350)

    Lp = np.power(10, np.random.uniform(-9, -6))# Diffusion length
    Ln = np.power(10, np.random.uniform(-9, -6)) # Diffusion length

    GaAs_window = material("GaAs")()
    GaAs_n = material("GaAs")(Nd = Nd, hole_diffusion_length=Ln)
    GaAs_p = material("GaAs")(Na = Na, electron_diffusion_length=Lp)

    n_width = np.random.uniform(500, 1000)*1e-9
    p_width = np.random.uniform(3000, 5000)*1e-9
    window_width = np.random.uniform(25, 200)*1e-9

    test_junc  = Junction([Layer(window_width, GaAs_window, role="window"),
                           Layer(n_width, GaAs_n,role="emitter"),
                           Layer(p_width, GaAs_p, role="base")])

    id_top, id_bottom, pRegion, nRegion, iRegion, pn_or_np = identify_layers(test_junc)
    xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd_c, Na_c, ni, es = identify_parameters(test_junc, options.T, pRegion, nRegion, iRegion)

    ni_expect = GaAs_n.ni

    assert [id_top, id_bottom] == approx([1, 2])
    assert pn_or_np == 'np'
    assert [xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd_c, Na_c, ni, es] == approx([n_width, p_width, 0, 0, 0, Lp, Ln, GaAs_n.hole_mobility * kb * options.T / q,
                                    GaAs_p.electron_mobility * kb * options.T / q, Nd, Na, ni_expect, GaAs_n.permittivity])




def test_process_junction_pn():
    from solcore.analytic_solar_cells.depletion_approximation import identify_layers, identify_parameters
    from solcore import material
    from solcore.structure import Layer, Junction
    from solcore.state import State

    Na = np.power(10, np.random.uniform(22, 25))
    Nd = np.power(10, np.random.uniform(23, 26))

    options = State()
    options.T = np.random.uniform(250, 350)

    Lp = np.power(10, np.random.uniform(-9, -6))  # Diffusion length
    Ln = np.power(10, np.random.uniform(-9, -6))  # Diffusion length

    GaAs_n = material("GaAs")(Nd=Nd, hole_diffusion_length=Ln)
    GaAs_p = material("GaAs")(Na=Na, electron_diffusion_length=Lp)

    p_width = np.random.uniform(500, 1000)*1e-9
    n_width = np.random.uniform(3000, 5000)*1e-9

    test_junc = Junction([Layer(p_width, GaAs_p, role="emitter"), Layer(n_width, GaAs_n, role="base")])

    id_top, id_bottom, pRegion, nRegion, iRegion, pn_or_np = identify_layers(test_junc)
    xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd_c, Na_c, ni, es = identify_parameters(test_junc, options.T, pRegion, nRegion, iRegion)

    ni_expect = GaAs_n.ni

    assert [id_top, id_bottom] == approx([0, 1])
    assert pn_or_np == 'pn'
    assert [xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd_c, Na_c, ni, es] == approx([n_width, p_width, 0, 0, 0, Lp, Ln, GaAs_n.hole_mobility * kb * options.T / q,
                                    GaAs_p.electron_mobility * kb * options.T / q, Nd, Na, ni_expect, GaAs_n.permittivity])


def test_process_junction_nip():
    from solcore.analytic_solar_cells.depletion_approximation import identify_layers, identify_parameters
    from solcore import material
    from solcore.structure import Layer, Junction
    from solcore.state import State

    Na = np.power(10, np.random.uniform(23, 26))
    Nd = np.power(10, np.random.uniform(22, 25))

    options = State()
    options.T = np.random.uniform(250, 350)

    Lp = np.power(10, np.random.uniform(-9, -6)) # Diffusion length
    Ln = np.power(10, np.random.uniform(-9, -6)) # Diffusion length

    GaAs_n = material("GaAs")(Nd = Nd, hole_diffusion_length=Ln)
    GaAs_p = material("GaAs")(Na = Na, electron_diffusion_length=Lp)
    GaAs_i = material("GaAs")()

    n_width = np.random.uniform(500, 1000)*1e-9
    p_width = np.random.uniform(3000, 5000)*1e-9
    i_width = np.random.uniform(100, 300)*1e-9

    test_junc  = Junction([Layer(n_width, GaAs_n,role="emitter"),
                           Layer(i_width, GaAs_i, role="intrinsic"),
                           Layer(p_width, GaAs_p, role="base")])

    id_top, id_bottom, pRegion, nRegion, iRegion, pn_or_np = identify_layers(test_junc)
    xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd_c, Na_c, ni, es = identify_parameters(test_junc, options.T, pRegion, nRegion, iRegion)

    ni_expect = GaAs_n.ni

    assert [id_top, id_bottom] == approx([0, 2])
    assert pn_or_np == 'np'
    assert [xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd_c, Na_c, ni, es] == approx([n_width, p_width, i_width, 0, 0, Lp, Ln, GaAs_n.hole_mobility * kb * options.T / q,
                                    GaAs_p.electron_mobility * kb * options.T / q, Nd, Na, ni_expect, GaAs_n.permittivity])



def test_process_junction_pin():
    from solcore.analytic_solar_cells.depletion_approximation import identify_layers, identify_parameters
    from solcore import material
    from solcore.structure import Layer, Junction
    from solcore.state import State

    Na = np.power(10, np.random.uniform(22, 25))
    Nd = np.power(10, np.random.uniform(23, 26))

    options = State()
    options.T = np.random.uniform(250, 350)

    Lp = np.power(10, np.random.uniform(-9, -6))  # Diffusion length
    Ln = np.power(10, np.random.uniform(-9, -6)) # Diffusion length

    GaAs_n = material("GaAs")(Nd=Nd, hole_diffusion_length=Ln)
    GaAs_p = material("GaAs")(Na=Na, electron_diffusion_length=Lp)
    GaAs_i = material("GaAs")()

    p_width = np.random.uniform(500, 1000)
    n_width = np.random.uniform(3000, 5000)
    i_width = np.random.uniform(100, 300)*1e-9

    test_junc = Junction([Layer(p_width, GaAs_p, role="emitter"),
                          Layer(i_width, GaAs_i, role="intrinsic"),
                          Layer(n_width, GaAs_n, role="base")])

    id_top, id_bottom, pRegion, nRegion, iRegion, pn_or_np = identify_layers(test_junc)
    xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd_c, Na_c, ni, es = identify_parameters(test_junc, options.T, pRegion, nRegion, iRegion)

    ni_expect = GaAs_n.ni

    assert [id_top, id_bottom] == approx([0, 2])
    assert pn_or_np == 'pn'
    assert [xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd_c, Na_c, ni, es] == approx([n_width, p_width, i_width, 0, 0, Lp, Ln, GaAs_n.hole_mobility * kb * options.T / q,
                                    GaAs_p.electron_mobility * kb * options.T / q, Nd, Na, ni_expect, GaAs_n.permittivity])


def test_process_junction_set_in_junction():
    from solcore.analytic_solar_cells.depletion_approximation import identify_layers, identify_parameters
    from solcore import material
    from solcore.structure import Layer, Junction
    from solcore.state import State

    options = State()
    options.T = np.random.uniform(250, 350)

    Lp = np.power(10, np.random.uniform(-9, -6)) # Diffusion length
    Ln = np.power(10, np.random.uniform(-9, -6)) # Diffusion length

    sn = np.power(10, np.random.uniform(0, 3))
    sp = np.power(10, np.random.uniform(0, 3))

    se = np.random.uniform(1, 20)*vacuum_permittivity

    GaAs_n = material("GaAs")()
    GaAs_p = material("GaAs")()
    GaAs_i = material("GaAs")()

    p_width = np.random.uniform(500, 1000)
    n_width = np.random.uniform(3000, 5000)
    i_width = np.random.uniform(100, 300)*1e-9

    mun = np.power(10, np.random.uniform(-5, 0))
    mup = np.power(10, np.random.uniform(-5, 0))

    Vbi = np.random.uniform(0, 3)

    test_junc = Junction([Layer(p_width, GaAs_p, role="emitter"),
                          Layer(i_width, GaAs_i, role="intrinsic"),
                          Layer(n_width, GaAs_n, role="base")],
                         sn = sn, sp = sp, permittivity = se,
                         ln = Ln, lp= Lp,
                         mup = mup, mun = mun, Vbi = Vbi)

    id_top, id_bottom, pRegion, nRegion, iRegion, pn_or_np = identify_layers(test_junc)
    xn, xp, xi, sn_c, sp_c, ln, lp, dn, dp, Nd_c, Na_c, ni, es = identify_parameters(test_junc, options.T, pRegion, nRegion, iRegion)

    ni_expect = GaAs_n.ni

    assert [id_top, id_bottom] == approx([0, 2])
    assert pn_or_np == 'pn'
    assert [xn, xp, xi, sn_c, sp_c, ln, lp, dn, dp, Nd_c, Na_c, ni, es] == approx([n_width, p_width, i_width, sn, sp, Ln, Lp, mun * kb * options.T / q,
                                    mup * kb * options.T / q, 1, 1, ni_expect, se])


def test_get_depletion_widths():
    from solcore.analytic_solar_cells.depletion_approximation import get_depletion_widths
    from solcore.structure import Junction

    xi  = np.power(10, np.random.uniform(-10, -6))

    Vbi = np.random.uniform(0, 3)
    es = np.random.uniform(1, 20)*vacuum_permittivity
    Na = np.power(10, np.random.uniform(22, 25))
    Nd = np.power(10, np.random.uniform(22, 25))

    V = np.linspace(-6, 4, 20)
    V = np.where(V < Vbi - 0.001, V, Vbi - 0.001)

    test_junc = Junction()

    wn_e = (-xi + np.sqrt(xi ** 2 + 2. * es * (Vbi - V) / q * (1 / Na + 1 / Nd))) / (1 + Nd / Na)
    wp_e = (-xi + np.sqrt(xi ** 2 + 2. * es * (Vbi - V) / q * (1 / Na + 1 / Nd))) / (1 + Na / Nd)

    wn_r, wp_r = get_depletion_widths(test_junc, es, Vbi, V, Na, Nd, xi)

    assert wn_r == approx(wn_e)
    assert wp_r == approx(wp_e)


def test_get_depletion_widths_onesided():
    from solcore.analytic_solar_cells.depletion_approximation import get_depletion_widths
    from solcore.structure import Junction

    xi  = np.power(10, np.random.uniform(-10, -6))

    Vbi = np.random.uniform(0, 3)
    es = np.random.uniform(1, 20)*vacuum_permittivity
    Na = np.power(10, np.random.uniform(22, 25))
    Nd = np.power(10, np.random.uniform(22, 25))

    V = np.linspace(-6, 4, 20)
    V = np.where(V < Vbi - 0.001, V, Vbi - 0.001)

    test_junc = Junction(depletion_approximation="one-sided abrupt")

    wn_e = np.sqrt(2 * es * (Vbi - V) / (q * Nd))
    wp_e = np.sqrt(2 * es * (Vbi - V) / (q * Na))

    wn_r, wp_r = get_depletion_widths(test_junc, es, Vbi, V, Na, Nd, xi)

    assert wn_r == approx(wn_e)
    assert wp_r == approx(wp_e)


def test_get_depletion_widths_set_in_junction():
    from solcore.analytic_solar_cells.depletion_approximation import get_depletion_widths
    from solcore.structure import Junction

    wn = np.random.uniform(1,100)
    wp = np.random.uniform(1,100)
    test_junc = Junction(wn=wn, wp=wp)

    wn_r, wp_r = get_depletion_widths(test_junc, 0, 0, 0, 0, 0, 0)

    assert wn_r == approx(wn)
    assert wp_r == approx(wp)


def test_dark_iv_depletion_pn(pn_junction):
    from solcore.analytic_solar_cells.depletion_approximation import iv_depletion, get_depletion_widths, get_j_dark, get_Jsrh, identify_layers, identify_parameters
    from scipy.interpolate import interp1d

    test_junc, options = pn_junction
    options.light_iv = False
    T = options.T

    id_top, id_bottom, pRegion, nRegion, iRegion, pn_or_np = identify_layers(test_junc[0])
    xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd, Na, ni, es = identify_parameters(test_junc[0], T, pRegion, nRegion, iRegion)

    niSquared = ni**2

    kbT = kb * T

    Vbi = (kbT / q) * np.log(Nd * Na / niSquared)

    test_junc[0].voltage = options.internal_voltages

    V = np.where(test_junc[0].voltage < Vbi - 0.001, test_junc[0].voltage, Vbi - 0.001)

    wn, wp = get_depletion_widths(test_junc[0], es, Vbi, V, Na, Nd, xi)

    w = wn + wp + xi

    l_top, l_bottom = ln, lp
    x_top, x_bottom = xp, xn
    w_top, w_bottom = wp, wn
    s_top, s_bottom = sp, sn
    d_top, d_bottom = dp, dn
    min_top, min_bot = niSquared / Na, niSquared / Nd

    JtopDark = get_j_dark(x_top, w_top, l_top, s_top, d_top, V, min_top, T)
    JbotDark = get_j_dark(x_bottom, w_bottom, l_bottom, s_bottom, d_bottom, V, min_bot, T)

    JnDark, JpDark = JbotDark, JtopDark

    lifetime_n = ln ** 2 / dn
    lifetime_p = lp ** 2 / dp  # Jenny p163

    Jrec = get_Jsrh(ni, V, Vbi, lifetime_p, lifetime_n, w, kbT)

    J_sc_top = 0
    J_sc_bot = 0
    J_sc_scr = 0

    current = Jrec + JnDark + JpDark + V / 1e14- J_sc_top - J_sc_bot - J_sc_scr
    iv = interp1d(test_junc[0].voltage, current, kind='linear', bounds_error=False, assume_sorted=True,
                           fill_value=(current[0], current[-1]), copy=True)

    iv_depletion(test_junc[0], options)

    assert test_junc[0].iv(options.internal_voltages) == approx(iv(options.internal_voltages), nan_ok=True)


def test_dark_iv_depletion_np(np_junction):
    from solcore.analytic_solar_cells.depletion_approximation import iv_depletion, get_depletion_widths, get_j_dark, get_Jsrh, identify_layers, identify_parameters
    from scipy.interpolate import interp1d

    test_junc, options = np_junction
    options.light_iv = False
    T = options.T

    test_junc[0].voltage = options.internal_voltages

    id_top, id_bottom, pRegion, nRegion, iRegion, pn_or_np = identify_layers(test_junc[0])
    xn, xp, xi, sn, sp, ln, lp, dn, dp, Nd, Na, ni, es = identify_parameters(test_junc[0], T, pRegion, nRegion, iRegion)

    niSquared = ni**2

    kbT = kb * T

    Vbi = (kbT / q) * np.log(Nd * Na / niSquared)

    V = np.where(test_junc[0].voltage < Vbi - 0.001, test_junc[0].voltage, Vbi - 0.001)

    wn, wp = get_depletion_widths(test_junc[0], es, Vbi, V, Na, Nd, xi)

    w = wn + wp + xi

    l_bottom, l_top = ln, lp
    x_bottom, x_top = xp, xn
    w_bottom, w_top = wp, wn
    s_bottom, s_top = sp, sn
    d_bottom, d_top = dp, dn
    min_bot, min_top = niSquared / Na, niSquared / Nd

    JtopDark = get_j_dark(x_top, w_top, l_top, s_top, d_top, V, min_top, T)
    JbotDark = get_j_dark(x_bottom, w_bottom, l_bottom, s_bottom, d_bottom, V, min_bot, T)

    JpDark, JnDark = JbotDark, JtopDark

    lifetime_n = ln ** 2 / dn
    lifetime_p = lp ** 2 / dp  # Jenny p163

    Jrec = get_Jsrh(ni, V, Vbi, lifetime_p, lifetime_n, w, kbT)

    J_sc_top = 0
    J_sc_bot = 0
    J_sc_scr = 0

    current = Jrec + JnDark + JpDark + V / 1e14- J_sc_top - J_sc_bot - J_sc_scr
    iv = interp1d(test_junc[0].voltage, current, kind='linear', bounds_error=False, assume_sorted=True,
                           fill_value=(current[0], current[-1]), copy=True)

    iv_depletion(test_junc[0], options)

    assert test_junc[0].iv(options.internal_voltages) == approx(iv(options.internal_voltages), nan_ok=True)


def test_qe_depletion_np(np_junction):

    from solcore.analytic_solar_cells import qe_depletion

    test_junc, options = np_junction

    wl = options.wavelength

    qe_depletion(test_junc[0], options)

    assert np.all(test_junc[0].eqe(wl) < 1)
    assert np.all(test_junc[0].eqe_emitter(wl) < 1)
    assert np.all(test_junc[0].eqe_base(wl) < 1)
    assert np.all(test_junc[0].eqe_scr(wl) < 1)
    assert np.all(test_junc[0].eqe(wl)[test_junc[0].eqe(wl) > 1e-3]
                  <= test_junc.absorbed[test_junc[0].eqe(wl) > 1e-3])
    assert np.all(test_junc[0].eqe_emitter(wl) + test_junc[0].eqe_base(wl) +
                  test_junc[0].eqe_scr(wl) == approx(test_junc[0].eqe(wl)))
    assert np.all(test_junc[0].iqe(wl) >= test_junc[0].eqe(wl))



def test_qe_depletion_pn(pn_junction):
    from solcore.analytic_solar_cells import qe_depletion

    test_junc, options = pn_junction

    wl = options.wavelength

    qe_depletion(test_junc[0], options)

    assert np.all(test_junc[0].eqe(wl) < 1)
    assert np.all(test_junc[0].eqe_emitter(wl) < 1)
    assert np.all(test_junc[0].eqe_base(wl) < 1)
    assert np.all(test_junc[0].eqe_scr(wl) < 1)
    assert np.all(test_junc[0].eqe(wl)[test_junc[0].eqe(wl) > 1e-3]
                  <= test_junc.absorbed[test_junc[0].eqe(wl) > 1e-3])
    assert np.all(test_junc[0].eqe_emitter(wl) + test_junc[0].eqe_base(wl) +
                  test_junc[0].eqe_scr(wl) == approx(test_junc[0].eqe(wl)))
    assert np.all(test_junc[0].iqe(wl) >= test_junc[0].eqe(wl))


def test_iv_depletion_np(np_junction):

    from solcore.analytic_solar_cells import iv_depletion

    test_junc, options = np_junction
    options.light_iv = True
    V = options.internal_voltages
    wl = options.wavelength

    iv_depletion(test_junc[0], options)

    wl_sp, ph = options.light_source.spectrum(output_units='photon_flux_per_m', x=wl)
    Jph = q*np.trapz(test_junc.absorbed*ph, wl)

    approx_Voc = V[np.argmin(abs(test_junc[0].iv(V)))]

    quadrant = (V > 0) * (V < approx_Voc)

    power = abs(test_junc[0].iv(V[quadrant])*V[quadrant])[:-1]


    assert abs(test_junc[0].iv(0)) <= Jph
    assert approx_Voc < test_junc[0][1].material.band_gap/q
    assert np.all(power < options.light_source.power_density)


def test_iv_depletion_pn(pn_junction):
    from solcore.analytic_solar_cells import iv_depletion

    test_junc, options = pn_junction
    options.light_iv = True
    V = options.internal_voltages
    wl = options.wavelength

    iv_depletion(test_junc[0], options)

    wl_sp, ph = options.light_source.spectrum(output_units='photon_flux_per_m', x=wl)
    Jph = q*np.trapz(test_junc.absorbed*ph, wl)

    approx_Voc = V[np.argmin(abs(test_junc[0].iv(V)))]

    quadrant = (V > 0) * (V < approx_Voc)

    power = abs(test_junc[0].iv(V[quadrant])*V[quadrant])[:-1]

    assert abs(test_junc[0].iv(0)) <= Jph
    assert approx_Voc < test_junc[0][1].material.band_gap/q
    assert np.all(power < options.light_source.power_density)




import numpy as np
from solcore.interpolate import interp1d
from solcore.structure import Junction, Layer
from solcore.constants import kb, q, hbar, electron_mass, vacuum_permittivity, pi
from solcore.science_tracker import science_reference
from solcore import eVnm, siUnits, convert

fs = 6.8e-5
Ts = 5762.0



#################
### Old Functions

def spectral_response_all_junctions(solar_cell, incident_light=None, energy=None, V=0, verbose=False):
    """ Calculates the spectral response of any number of junctions using analytical diode equations as described in
    J. Nelson's book "The Physics of Solar Cells" (2003). All parameters must be in SI units. It only works for
    homojunctions.

    :param solar_cell: A structure object with one or more Junction objects and a few optional parameters. Each junction
     is made of a sequence of *Layers* (with a thickness, a material and a role) and the surface recombination
     velocities for electrons and holes. Optional attributes for the structure are:
        - shading: (optional) Shading loses.
        - reflectivity: (optional) Function that calculates the reflectivity as a function of energy (in J).
    :param incident_light: (optional) 2D array containing the energies and the spectral power density (in photons m-2 J-1).
    :param energy: (optional) energies at which to calculate the QE. If not provided, the range of the incident ligth
    is used and if that is not defined, a "sensible" range is used.
    :param V: The voltage at which to perform the calculations (in V)
    :param verbose: If the information about the calculation must be printed (default = FALSE).
    :return: A dictionary containing, as a function of energy:

        - junctions: A list containing the QE data for each junction
        - transmitted: The transmitted power density
        - transmitted_fraction: The fraction of transmitted power
        - passive_loss: The passive losses (light absorbed in the encapsulants, AR coatings, window layers, etc)
        - reflected: reflected power density
        - reflected_fraction: Fraction of reflected power
        - e: The energy values
    """
    science_reference("Nelson pin spectral response",
                      "Jenny: (Nelson. The Physics of Solar Cells. Imperial College Press (2003))")

    # Get the energy range and incident spectrum. If they are not inputs, we create some sensible values.
    if energy is None:
        if incident_light is not None:
            energy = incident_light[0]
            bs = np.copy(incident_light[1])
        else:
            energy = siUnits(np.linspace(0.5, 3.5, 450), 'eV')
            bs = np.ones_like(energy)
    else:
        if incident_light is not None:
            bs = np.interp(energy, incident_light[0], incident_light[1])
        else:
            bs = np.ones_like(energy)

    bs_initial = np.copy(bs)

    # We include the shadowing losses
    if hasattr(solar_cell, 'shading'):
        bs *= (1 - solar_cell.shading)

    # And the reflexion losses
    if hasattr(solar_cell, 'reflectivity') and solar_cell.reflectivity is not None:
        ref = solar_cell.reflectivity(energy)
        bs *= (1 - ref)
        reflected = ref * bs_initial
    else:
        reflected = np.zeros_like(bs)

    # And now we perform the calculation, each junction at a time
    qe_result = []
    passive_loss = np.ones_like(bs)

    for layer_index, layer_object in enumerate(solar_cell):

        # Attenuation due to absorption in the AR coatings or any layer in the front that is not part of the
        # junction
        if type(layer_object) is Layer:
            bs = bs * np.exp(-layer_object.material.alphaE(energy) * layer_object.width)
            passive_loss *= np.exp(-layer_object.material.alphaE(energy) * layer_object.width)

        # For each junction, we calculate the spectral response
        elif type(layer_object) is Junction:

            # If there are window layers, passively absorbing light above the emitter, we attenuate the intensity
            idx = 0
            for junction_layer_object in layer_object:
                if junction_layer_object.role != 'emitter':
                    bs = bs * np.exp(-junction_layer_object.material.alphaE(energy) * junction_layer_object.width)
                    passive_loss *= np.exp(-junction_layer_object.material.alphaE(energy) * junction_layer_object.width)
                    idx += 1
                else:
                    break

            output = calculate_junction_sr(layer_object, energy, bs, bs_initial, V, printParameters=verbose)
            qe_result.append(output)

            # And we reduce the amount of light reaching the next junction
            for junction_layer_object in layer_object[idx:]:
                bs *= np.exp(-junction_layer_object.material.alphaE(energy) * junction_layer_object.width)

        else:
            raise ValueError("Strange layer-like object discovered in structure stack: {}".format(type(layer_object)))

    return {"junctions": qe_result,
            "transmitted": bs,
            "transmitted_fraction": bs / bs_initial,
            "passive_loss": 1 - passive_loss,
            "reflected": reflected,
            "reflected_fraction": reflected / bs_initial,
            "e": energy}


def calculate_junction_sr(junc, energies, bs, bs_initial, V, printParameters=False):
    """ Calculates the total quantum efficiency, the QE splitted by regions, photocurrent and other parameters for a
    given junction at a given voltage.

    :param junc: The junction object
    :param energies: The energies at which to perform the calculation
    :param bs: The spectral power density reaching the junction
    :param bsInitial: The initial power density
    :param V: The voltage at which to perform the calculations (not implemented, yet)
    :param printParameters: If a list of all parameters must be printed
    :return:
    """

    # First we have to figure out if we are talking about a PN, NP, PIN or NIP junction
    sn = 0 if not hasattr(junc, "sn") else junc.sn
    sp = 0 if not hasattr(junc, "sp") else junc.sp

    # First we search for the emitter and check if it is n-type or p-type
    idx = 0
    pn_or_np = 'pn'
    for layer in junc:
        if layer.role is not 'emitter':
            idx += 1
        else:
            Na = 0
            Nd = 0
            if hasattr(layer.material, 'Na'): Na = layer.material.Na
            if hasattr(layer.material, 'Nd'): Nd = layer.material.Nd
            if Na < Nd:
                pn_or_np = "np"
                nRegion = junc[idx]
            else:
                pRegion = junc[idx]
            break

    # Now we check for an intrinsic region and, if there is, for the base.
    if junc[idx + 1].role is 'intrinsic':
        iRegion = junc[idx + 1]

        if junc[idx + 2].role is 'base':
            if pn_or_np == "pn":
                nRegion = junc[idx + 2]
            else:
                pRegion = junc[idx + 2]
        else:
            raise RuntimeError('ERROR processing junctions: A layer following the "intrinsic" layer must be defined as '
                               '"base".')

    # If there is no intrinsic region, we check directly the base
    elif junc[idx + 1].role is 'base':
        if pn_or_np == "pn":
            nRegion = junc[idx + 1]
        else:
            pRegion = junc[idx + 1]
        iRegion = None

    else:
        raise RuntimeError('ERROR processing junctions: A layer following the "emitter" must be defined as "intrinsic"'
                           'or "base".')

    # With all regions identified, it's time to start doing calculations
    T = nRegion.material.T
    kbT = kb * T
    Egap = nRegion.material.band_gap

    xp = pRegion.width
    xn = nRegion.width
    xi = 0 if iRegion is None else iRegion.width

    # Now we have to get all the material parameters needed for the calculation
    if hasattr(junc, "dielectric_constant"):
        es = junc.dielectric_constant
    else:
        es = nRegion.material.permittivity * vacuum_permittivity  # equal for n and p.  I hope.

    # For the diffusion lenght, subscript n and p refer to the carriers, electrons and holes
    if hasattr(junc, "ln"):
        ln = junc.ln
    else:
        ln = pRegion.material.electron_diffusion_length

    if hasattr(junc, "lp"):
        lp = junc.lp
    else:
        lp = nRegion.material.hole_diffusion_length

    # For the diffusion coefficient, n and p refer to the regions, n side and p side. Yeah, it's confusing...
    if hasattr(junc, "mup"):
        dp = junc.mup * kb * T / q
    else:
        dp = pRegion.material.electron_mobility * kb * T / q

    if hasattr(junc, "mun"):
        dn = junc.mun * kb * T / q
    else:
        dn = nRegion.material.hole_mobility * kb * T / q

    # Effective masses and effective density of states
    mEff_h = nRegion.material.eff_mass_hh_z * electron_mass
    mEff_e = pRegion.material.eff_mass_electron * electron_mass

    Nv = 2 * (mEff_h * kb * T / (2 * pi * hbar ** 2)) ** 1.5  # Jenny p58
    Nc = 2 * (mEff_e * kb * T / (2 * pi * hbar ** 2)) ** 1.5
    niSquared = Nc * Nv * np.exp(-Egap / (kb * T))
    ni = np.sqrt(niSquared)

    Na = pRegion.material.Na
    Nd = nRegion.material.Nd
    Vbi = (kb * T / q) * np.log(Nd * Na / niSquared) if not hasattr(junc, "Vbi") else junc.Vbi  # Jenny p146

    # And now we account for the possible applied voltage, which can be, at most, equal to Vbi
    V = min(Vbi, V)
    Vbi = Vbi - V

    # It's time to calculate the depletion widths
    if not hasattr(junc, "wp") or not hasattr(junc, "wn"):

        if hasattr(junc, "depletion_approximation") and junc.depletion_approximation == "one-sided abrupt":
            print("using one-sided abrupt junction approximation for depletion width")
            science_reference("Sze abrupt junction approximation",
                              "Sze: The Physics of Semiconductor Devices, 2nd edition, John Wiley & Sons, Inc (2007)")
            wp = np.sqrt(2 * es * Vbi / (q * Na))
            wn = np.sqrt(2 * es * Vbi / (q * Nd))

        else:
            wn = (-xi + np.sqrt(xi ** 2 + 2. * es * Vbi / q * (1 / Na + 1 / Nd))) / (1 + Nd / Na)
            wp = (-xi + np.sqrt(xi ** 2 + 2. * es * Vbi / q * (1 / Na + 1 / Nd))) / (1 + Na / Nd)

    wn = wn if not hasattr(junc, "wn") else junc.wn
    wp = wp if not hasattr(junc, "wp") else junc.wp

    # we have an array of alpha values that needs to be interpolated to the right energies
    alphaN = nRegion.material.alphaE(energies)  # create numpy array at right energies.
    alphaP = pRegion.material.alphaE(energies)  # create numpy array at right energies.
    alphaI = iRegion.material.alphaE(energies) if iRegion else 0
    depleted_width = wn + wp + xi
    bs_incident_on_top = bs

    # Now it is time to calculate currents
    if pn_or_np == "pn":
        bs_incident_on_bottom = bs * np.exp(-alphaP * xp - alphaN * wn - xi * alphaI)
        bs_incident_on_depleted = bs * np.exp(-alphaP * (xp - wp))
        alphaTop = alphaP
        alphaBottom = alphaN

        l_top, l_bottom = ln, lp
        x_top, x_bottom = xp, xn
        w_top, w_bottom = wp, wn
        s_top, s_bottom = sp, sn
        d_top, d_bottom = dp, dn
        min_top, min_bot = niSquared / Na, niSquared / Nd
    else:
        bs_incident_on_bottom = bs * np.exp(-alphaN * xn - alphaP * wp - xi * alphaI)
        bs_incident_on_depleted = bs * np.exp(-alphaN * (xn - wn))
        alphaTop = alphaN
        alphaBottom = alphaP

        l_bottom, l_top = ln, lp
        x_bottom, x_top = xp, xn
        w_bottom, w_top = wp, wn
        s_bottom, s_top = sp, sn
        d_bottom, d_top = dp, dn
        min_bot, min_top = niSquared / Na, niSquared / Nd

    j_top, JtopDark = get_j_top(x_top, w_top, l_top, s_top, d_top, alphaTop, bs_incident_on_top, V, min_top, T)
    j_bottom, JbotDark = get_j_bot(x_bottom, w_bottom, l_bottom, s_bottom, d_bottom, alphaBottom, bs_incident_on_bottom,
                                   V, min_bot, T)

    jgen = q * bs_incident_on_depleted * (1 - np.exp(-alphaI * xi - alphaN * wn - alphaP * wp))  # jgen. Jenny, p. 159

    # hereby we define the subscripts to refer to the layer in which the current is generated:
    if pn_or_np == "pn":
        jn, jp = j_bottom, j_top
        JnDark, JpDark = JbotDark, JtopDark
    else:
        jp, jn = j_bottom, j_top
        JpDark, JnDark = JbotDark, JtopDark

    # These might not be the right lifetimes. Actually, they are not as they include all recombination processes, not
    # just SRH recombination, which is what the equation in Jenny, p159 refers to. Let´ leave them, for now.
    lifetime_n = ln ** 2 / dn
    lifetime_p = lp ** 2 / dp  # Jenny p163

    # Jrec. Jenny, p.159. Note capital J. This does not need integrating over energies
    Jrec = q * ni * (wn + wp + xi) / np.sqrt(lifetime_n * lifetime_p) * np.sinh(q * V / (2 * kbT)) / (
        q * Vbi / kbT) * pi

    # jgen = q* bs*(1 - exp(-depleted_width*alpha))*exp(-(xn-wn)*alpha);
    nDepletionCharge = wn * Nd * q
    pDepletionCharge = wp * Na * q

    Vbi2 = (0.5 * (wn + wp) + xi) * pDepletionCharge / es

    good_indeces = np.isfinite(jn) * np.isfinite(jp) * np.isfinite(jgen)

    energies = energies[good_indeces]
    jn = jn[good_indeces]
    jp = jp[good_indeces]
    jgen = jgen[good_indeces]

    # jn[jn < 0] = 0
    # jp[jp < 0] = 0
    # jgen[jgen < 0] = 0

    bs_initial = bs_initial[good_indeces]

    Jn = np.trapz(y=jn, x=energies)
    Jp = np.trapz(y=jp, x=energies)
    Jgen = np.trapz(y=jgen, x=energies)

    if printParameters:
        jSum = list((jn + jp + jgen) / bs_initial / q)
        peakQE = max(jSum)
        BandgapEV = convert(Egap, "J", "eV")
        peakQEE = convert(energies[jSum.index(peakQE)], "J", "eV")

        parameterDictionary = {
            "typee": "PIN" if iRegion is not None else "PN",
            "Na": Na, "Nd": Nd,
            "mEff_e": mEff_e, "mEff_h": mEff_h,
            "dp": dp, "dn": dn, "T": T, "es": es, "Vbi": Vbi,
            "xpUm": convert(xp, "m", "um"),
            "xnUm": convert(xn, "m", "um"),
            "BandgapEV": BandgapEV,
            "BandgapNM": eVnm(BandgapEV),
            "pMatrialString": str(pRegion.material),
            "nMatrialString": str(nRegion.material),
            "relativeEffectiveMassE": mEff_e / electron_mass,
            "relativeEffectiveMassH": mEff_h / electron_mass,
            "wnNm": convert(wn, "m", "nm"),
            "wpNm": convert(wp, "m", "nm"),
            "NaCM": convert(Na, "m-3", "cm-3"),
            "NdCM": convert(Nd, "m-3", "cm-3"),
            "permittivity": es / vacuum_permittivity,
            "peakQE": peakQE * 100,
            "peakQEE": peakQEE,
            "peakQENM": eVnm(peakQEE),
            "lpum": convert(lp, "m", "um"),
            "lnum": convert(ln, "m", "um"),
            "nQ": convert(nDepletionCharge, "m-2", "cm-2"),
            "pQ": convert(pDepletionCharge, "m-2", "cm-2"),
            "pDepletionCharge": pDepletionCharge,
            "nDepletionCharge": nDepletionCharge,
            "fieldMax": pDepletionCharge / (es),
            "iRegionString": "",
            "Vbi2": Vbi2,
            "ni": ni
        }
        if iRegion is not None:
            parameterDictionary["iRegionString"] = """
| i region: {xiUm:.2f} um {iMatrialString}
|   field: {fieldMax:.3e} V/m""".format(**{"iMatrialString": str(iRegion.material), "xiUm": convert(xi, "m", "um"),
                                           "fieldMax": pDepletionCharge / (es), })

        print("""\n
| Calculating {typee} QE. Active Parameters:
|
| p region: {xpUm:.2f} um {pMatrialString}
|   Na = {Na:.3e} m-3 ({NaCM:.3e} cm-3)
|   minority carrier (electron) diffusion length: {lpum:.3f} um
|   minority carrier (electron) effective mass: {relativeEffectiveMassE:.3f} (relative) {mEff_e:.3e} kg (absolute)
|   minority carrier (electron) diffusivity = {dp:.3e} m2/s
|   depletion width: {wpNm:.3f} nm
|   charge in depletion region: {pDepletionCharge:.3e} C m-2 ({pQ:.3e} C cm-2){iRegionString}
| n region: {xnUm:.2f} um {nMatrialString}
|   Nd = {Nd:.3e} m-3 ({NdCM:.3e} cm-3)
|   minority carrier (hole) diffusion length: {lnum:.3f} um
|   minority carrier (hole) effective mass: {relativeEffectiveMassH:.3f} (relative) {mEff_h:.3e} kg (absolute)
|   minority carrier (hole) diffusivity = {dn:.3e} m2/s
|   depletion width: {wnNm:.3f} nm
|   charge in depletion region: {nDepletionCharge:.3e} C m-2 ({nQ:.3e} C cm-2)
| Bandgap: {BandgapEV:.3f} eV ({BandgapNM:.2f} nm)
| Temperature: {T:.2f} K
| permittivity: {permittivity:.3f} (relative) {es:.3e} A s V-1 m-1 (absolute)
| built-in Voltage: {Vbi:.3f} V
| peak field: {fieldMax:.3e} V/m
| ni: {ni:e}
| Result:
| \tPeak QE = {peakQE:.1f} % at {peakQEE:.3f} eV ({peakQENM:.2f} nm)""".format(**parameterDictionary))

    return {
        "qe_n": jn / q / bs_initial,
        "qe_p": jp / q / bs_initial,
        "qe_scr": jgen / q / bs_initial,
        "qe_tot": (jn + jp + jgen) / q / bs_initial,
        "Jn_sc": Jn,
        "Jp_sc": Jp,
        "Jscr_sc": Jgen,
        "Jn_dif": JnDark,
        "Jp_dif": JpDark,
        "Jscr_srh": Jrec,
        "J": (Jn + Jp + Jgen - Jrec - JnDark - JpDark),
        "e": energies,
        "Temporary locals dictionary for radiative efficiency": locals()
    }


def get_j_top(x, w, l, s, d, alpha, bs, V, minority, T):
    # We calculate some fractions
    harg = (x - w) / l
    sinh_harg = np.sinh(harg)
    cosh_harg = np.cosh(harg)
    lsod = (l * s) / d

    # And then we are ready to calculate the different currents
    # Missing the voltage dependent part of these equations.
    # They should be 6.34 and 6.39, not 6.62 and 6.63
    j_top_light = (q * bs * alpha * l) / (alpha ** 2 * l ** 2 - 1) * \
                  ((lsod + alpha * l - np.exp(-alpha * (x - w)) * (lsod * cosh_harg + sinh_harg)) /
                   (lsod * sinh_harg + cosh_harg)
                   - alpha * l * np.exp(-alpha * (x - w)))

    J_top_dark = (q * d * minority / l) * (np.exp(q * V / kb / T) - 1) * \
                 ((lsod * cosh_harg + sinh_harg) / (lsod * sinh_harg + cosh_harg))

    return j_top_light, J_top_dark


def get_j_bot(x, w, l, s, d, alpha, bs, V, minority, T):
    # We calculate some fractions
    harg = (x - w) / l
    cosh_harg = np.cosh(harg)
    sinh_harg = np.sinh(harg)
    lsod = (l * s) / d

    # And then we are ready to calculate the different currents
    # Missing the voltage dependent part of these equations.
    # They should be 6.34 and 6.39, not 6.62 and 6.63

    j_bottom_light = (q * bs * alpha * l) / (alpha ** 2 * l ** 2 - 1) * \
                     (l * alpha -
                      (lsod * cosh_harg + sinh_harg - (lsod - l * alpha) * np.exp(-alpha * (x - w))) /
                      (cosh_harg + lsod * sinh_harg))

    J_bottom_dark = (q * d * minority / l) * (np.exp(q * V / kb / T) - 1) * \
                    ((lsod * cosh_harg + sinh_harg) / (lsod * sinh_harg + cosh_harg))

    return j_bottom_light, J_bottom_dark








import solcore.science_tracker as st

st.science_reference('GaP refractive index', 'S. Adachi. Optical dispersion relations for GaP, GaAs, GaSb, InP, InAs, InSb, AlxGa1−xAs, and In1−xGaxAsyP1−y, J. Appl. Phys. 66, 6030-6040 (1989)')



""" Dielectric constant model tests
"""
from pytest import approx
import numpy as np
import random

from solcore.absorption_calculator.dielectric_constant_models import (
    Cauchy,
    Drude,
    Gauss,
    Lorentz,
    Poles,
    PolySegment,
)
from solcore.absorption_calculator.dielectric_constant_models import (
    DielectricConstantModel,
)
from solcore.interpolate import interp1d


def test_poles():
    A = random.uniform(0, 10)
    Ec = random.uniform(0, 10)
    pole1 = Poles(A, Ec)
    assert pole1.A == A
    assert pole1.Ec == Ec

    pole2 = Poles()
    assert pole2.dielectric(random.uniform(0.1, 10)) == 0

    pole3 = Poles(3, 2)
    assert pole3.dielectric(1240) == 4


def test_lorentz():
    An = random.uniform(0, 10)
    En = random.uniform(0, 10)
    Brn = random.uniform(0, 10)
    lorentz1 = Lorentz(An, En, Brn)
    assert lorentz1.An == An
    assert lorentz1.En == En
    assert lorentz1.Brn == Brn

    lorentz2 = Lorentz(3, 2, 3)
    assert lorentz2.dielectric(1240) == (2 + 2j)


def test_gauss():
    A = random.uniform(0, 10)
    Ec = random.uniform(0, 10)
    Br = random.uniform(0, 10)
    gauss1 = Gauss(A, Ec, Br)
    assert gauss1.A == A
    assert gauss1.Ec == Ec
    assert gauss1.Br == Br

    gauss2 = Gauss(random.uniform(0.1, 10), 0, random.uniform(0.1, 10))
    assert gauss2.dielectric(random.randint(1, 100000)) == 0

    gauss3 = Gauss(A=2.8468, Ec=0.1299, Br=0.0111)
    assert gauss3.dielectric(1240) == approx(-0.00282958252877878)

    gauss4 = Gauss(A=7.192, Ec=0.058331, Br=0.01682)
    assert gauss4.dielectric(1240) == approx(-0.004798822090878001)

    gauss5 = Gauss(A=1.9737, Ec=0.13991, Br=0.02144)
    assert gauss5.dielectric(1240) == approx(-0.004093225570635009)


def test_drude():
    An = random.uniform(0, 10)
    Brn = random.uniform(0, 10)
    drude1 = Drude(An, Brn)
    assert drude1.An == An
    assert drude1.Brn == Brn

    drude2 = Drude(random.uniform(0.1, 10), 0)
    assert drude2.dielectric(random.uniform(0.1, 10)) == 0

    drude3 = Drude(2, 1)
    assert drude3.dielectric(1240) == (-1 + 1j)


def test_cauchy():
    An = random.uniform(0, 10)
    Bn = random.uniform(0, 10)
    Cn = random.uniform(0, 10)
    Ak = random.uniform(0, 10)
    Bk = random.uniform(0, 10)
    Ck = random.uniform(0, 10)
    cauchy1 = Cauchy(An, Bn, Cn, Ak, Bk, Ck)
    assert cauchy1.An == An
    assert cauchy1.Bn == Bn
    assert cauchy1.Cn == Cn
    assert cauchy1.Ak == Ak
    assert cauchy1.Bk == Bk
    assert cauchy1.Ck == Ck

    cauchy2 = Cauchy()
    assert cauchy2.dielectric(random.randint(0, 100000)) == 0

    cauchy3 = Cauchy(1, 2, 3, 1, np.log(2), 0.24)
    assert cauchy3.dielectric(1000) == (32 + 24j)


def test_poly_segment():
    energy = np.array([1, 100])
    e2 = np.array([1, 100])
    poly1 = PolySegment(energy, e2)
    assert all([input == output for input, output in zip(energy, poly1.energy)])
    assert all([input == output for input, output in zip(e2, poly1.e2)])
    assert isinstance(poly1.epsi2, interp1d)
    assert poly1.var == 2
    interp_value = random.uniform(1, 100)
    assert poly1.epsi2(interp_value) == interp_value  # linear interpolation
    assert poly1.dielectric(1240) == approx(1.008422827635958 + 1j)


def test_dielectric_constant_model(mocker):
    mocker.patch("builtins.input", return_value="Y")
    cauchy = Cauchy()
    model = DielectricConstantModel(e_inf=0, oscillators=[cauchy])
    assert model.dielectric_constants(1000) == 0
    model.add_oscillator("drude", An=2, Brn=1)
    assert model.dielectric_constants(1240) == (-1 + 1j)
    model.add_oscillator("lorentz", An=3, En=2, Brn=3)
    assert model.dielectric_constants(1240) == (1 + 3j)
    model.remove_oscillator(2)
    assert model.dielectric_constants(1240) == (2 + 2j)



from solcore.structure import Layer, Junction, TunnelJunction
import solcore.analytic_solar_cells as ASC

import numpy as np
import types
from scipy.interpolate import interp1d
#import matplotlib.pyplot as plt


def solve_beer_lambert(solar_cell, options):
    """ Calculates the reflection, transmission and absorption of a solar cell object using the Beer-Lambert law. Reflection is not really calculated and needs to be provided externally, otherwise it is assumed to be zero.

    :param solar_cell:
    :param options:
    :return:
    """
    wl_m = options.wavelength
    solar_cell.wavelength = options.wavelength

    fraction = np.ones(wl_m.shape)

    # We include the shadowing losses
    if hasattr(solar_cell, 'shading'):
        fraction *= (1 - solar_cell.shading)

    # And the reflexion losses
    if hasattr(solar_cell, 'reflectivity') and solar_cell.reflectivity is not None:
        solar_cell.reflected = solar_cell.reflectivity(wl_m)
        fraction *= (1 - solar_cell.reflected)
    else:
        solar_cell.reflected = np.zeros(fraction.shape)

    # Now we calculate the absorbed and transmitted light. We first get all the relevant parameters from the objects
    widths = []
    alphas = []
    for j, layer_object in enumerate(solar_cell):

        # Attenuation due to absorption in the AR coatings or any layer in the front that is not part of the junction
        if type(layer_object) is Layer:
            widths.append(layer_object.width)
            alphas.append(layer_object.material.alpha(wl_m))

        # For each Tunnel junctions will have, at most, a resistance an some layers absorbing light.
        elif type(layer_object) is TunnelJunction:
            for i, layer in enumerate(layer_object):
                widths.append(layer.width)
                alphas.append(layer.material.alpha(wl_m))

        # For each junction, and layer within the junction, we get the absorption coeficient and the layer width.
        elif type(layer_object) is Junction:

            kind = solar_cell[j].kind if hasattr(solar_cell[j], 'kind') else None

            if kind == '2D':
                # If the junction has a Jsc or EQE already defined, we ignore that junction in the optical calculation
                if hasattr(solar_cell[j], 'jsc') or hasattr(solar_cell[j], 'eqe'):
                    print('Warning: A junction of kind "2D" found. Junction ignored in the optics calculation!')

                    w = layer_object.width

                    def alf(x):
                        return 0.0*x

                    solar_cell[j].alpha = alf
                    solar_cell[j].reflected = interp1d(wl_m, solar_cell.reflected, bounds_error=False,
                                                       fill_value=(0, 0))

                    widths.append(w)
                    alphas.append(alf(wl_m))

                # Otherwise, we try to treat is as a DB junction from the optical point of view
                else:
                    ASC.absorptance_detailed_balance(solar_cell[j])
                    w = layer_object.width

                    def alf(x):
                        return -1 / w * np.log(np.maximum(1 - layer_object.absorptance(x), 1e-3))

                    solar_cell[j].alpha = alf
                    solar_cell[j].reflected = interp1d(wl_m, solar_cell.reflected, bounds_error=False,
                                                       fill_value=(0, 0))

                    widths.append(w)
                    alphas.append(alf(wl_m))

            elif kind == 'DB':
                ASC.absorptance_detailed_balance(solar_cell[j])
                w = layer_object.width

                def alf(x):
                    return -1 / w * np.log(np.maximum(1 - layer_object.absorptance(x), 1e-3))

                solar_cell[j].alpha = alf
                solar_cell[j].reflected = interp1d(wl_m, solar_cell.reflected, bounds_error=False, fill_value=(0, 0))

                widths.append(w)
                alphas.append(alf(wl_m))

            else:
                for i, layer in enumerate(layer_object):
                    widths.append(layer.width)
                    alphas.append(layer.material.alpha(wl_m))

    # With all this information, we are ready to calculate the absorbed light
    diff_absorption, transmitted, all_absorbed = calculate_absorption_beer_lambert(widths, alphas, fraction)

    # Each building block (layer or junction) needs to have access to the absorbed light in its region.
    # We update each object with that information.
    I_0 = fraction
    for j in range(len(solar_cell)):
        solar_cell[j].diff_absorption = diff_absorption
        solar_cell[j].absorbed = types.MethodType(absorbed, solar_cell[j])

        # total absorption at each wavelength, per layer
        layer_positions = options.position[(options.position >= solar_cell[j].offset) & (
                options.position < solar_cell[j].offset + solar_cell[j].width)]
        layer_positions = layer_positions - np.min(layer_positions)
        solar_cell[j].layer_absorption = np.trapz(solar_cell[j].absorbed(layer_positions), layer_positions, axis=0)

    solar_cell.transmitted = transmitted
    solar_cell.absorbed = all_absorbed


def absorbed(self, z):
    out = self.diff_absorption(self.offset + z).T * (z < self.width)
    return out.T


def calculate_absorption_beer_lambert(widths, alphas, fraction):
    # Number of spectral elements
    N = len(alphas[0])

    cum_widths = np.cumsum([0] + widths)
    widths = np.array(widths)

    # At any given position, the absorption per unit length is alpha * exp(-alpha*z) but we have to remove all light
    # absorbed above it:
    OD = [np.zeros(N)]
    for i in range(len(widths)):
        OD.append(OD[i] + alphas[i] * widths[i])

    # After getting al the alphas and widths, we need to create a function that takes as arguments the depth z and
    # returns the differential fraction of absorbed light at that position.
    def diff_absorption(z):
        # First we find to which layer a given z belong
        idx = cum_widths.searchsorted(z) - 1
        idx = np.maximum(idx, 0)

        # Now the distance to the begining of that layer
        z_local = z - cum_widths[idx]

        # And finally, we calculate a 2D array of the absorption per unit length vs position and wavelength
        output = np.zeros((len(z), N))

        inside = np.sum(z < cum_widths[-1])
        for k in range(inside):
            loc = idx[k]
            output[k] = alphas[loc] * np.exp(-OD[loc] - alphas[loc] * z_local[k])

        return output * fraction

    all_absorbed = (1 - np.exp(-OD[-1])) * fraction
    trans = np.exp(-OD[-1]) * fraction

    return diff_absorption, trans, all_absorbed


def calculate_absorptance_PDD_and_DA(junction, wl, fraction):
    widths = []
    alphas = []
    for i, layer in enumerate(junction):
        try:
            widths.append(layer.width)
            alphas.append(layer.material.alpha(wl))
        except Exception as err:
            # We are, most likely, in the presence of a QW structure. It needs to be already processed,
            # otherwise the calculation will fail.
            # We skip this part, for now.
            print(err.args)
            raise

    cum_widths = np.cumsum([0] + widths[:-1])
    widths = np.array(widths)

    # At any given position, the absorption per unit length is alpha * exp(alpha*z) but we have to remove all light
    # absorbed above it:
    OD = [np.zeros(alphas[0].shape)]
    for i in range(len(widths) - 1):
        OD.append(OD[i] + alphas[i] * widths[i])

    # After getting al the alphas and widths, we need to create a function that takes as arguments the depth z and
    # returns the differential fraction of absorbed light at that position.
    def diff_absorption(z):

        # First we find to which layer a given z belong
        idx = cum_widths.searchsorted(z) - 1
        idx = np.maximum(idx, 0)

        # Now the distance to the begining of that layer
        z_local = z - cum_widths[idx]

        # If we go beyond the limit of the function, the absorption must be zero, so:
        too_far = 1 * (z < (cum_widths[-1] + widths[-1]))

        # And finally, we calculate a 2D array of the absorption per unit lenght vs position and wavelength
        output = np.zeros((len(z), len(wl)))
        for i in range(len(z) - 1):
            loc = idx[i]
            output[i] = alphas[loc] * np.exp(-OD[loc] - alphas[loc] * z_local[i])
            output[i] = output[i] * too_far[i]

        return output * fraction

    trans = np.exp(-OD[-1] - alphas[-1] * widths[-1])

    return diff_absorption, trans


def calculate_absorptance_DB(junction, wl, fraction):
    # First, we calculate the absorptance of the cell
    ASC.absorptance_detailed_balance(junction)

    def absorption(new_wl):
        # The fraction of light absorbed in the junction
        f = np.interp(new_wl, wl, fraction, left=fraction[0], right=fraction[-1])
        return junction.absorptance(new_wl) * f

    trans = absorption(wl)

    return absorption, trans



import numpy as np
import os
import shutil
import subprocess
import tempfile
import platform
import solcore
from datetime import datetime


class SmartsSolverError(Exception):
    pass


smarts = solcore.config.smarts()


system = platform.system()
if system == 'Windows':
    extension = '.exe'
elif system == 'Linux':
    extension = ''
else:
    extension = '.command'

executable = os.path.join(smarts, "smarts295" + extension)

error_msg = """ERROR: SMARTS location not correctly configured or SMARTS executable not working.

SMARTS can be obtained free of charge from the NREL webpage:

http://www.nrel.gov/rredc/smarts/

You might need to re-compile the Fortran code as current 64 bit CPUs are not supported by the shipped binaries.
"""


def skipper(fname):
    with open(fname) as fin:
        no_comments = (line for line in fin if "   " in line)
        next(no_comments, None)  # skip header
        next(no_comments, None)  # skip header
        for row in no_comments:
            yield row


def calculate_spectrum_smarts(smarts_file_contents=None, filename='smarts295', target_directory=None):
    """

    :param smarts_file_contents:
    :param filename:
    :param target_directory:
    :return:
    """
    if not smarts:
        raise SmartsSolverError(f"Smarts installation not found in {smarts}")

    if smarts_file_contents is None:
        smarts_file_contents = build_smarts_file(**get_default_smarts_object())
    else:
        smarts_file_contents = build_smarts_file(**smarts_file_contents)

    if target_directory is not None:
        target_directory = target_directory.rstrip('\/')
        assert os.access(target_directory, os.W_OK), 'ERROR: Target folder for smarts output does not exists ' \
                                                     'or is not writable.'

    data = []
    # We use a temp directory to store temporary the data.
    # If needed, we'll copy it later to the target directory.
    with tempfile.TemporaryDirectory(prefix="tmp", suffix="_sc3SMARTS") as working_directory:

        ext_file = os.path.join(working_directory, "{}.ext.txt".format(filename))
        inp_file = os.path.join(working_directory, "{}.inp.txt".format(filename))
        out_file = os.path.join(working_directory, "{}.out.txt".format(filename))
        scn_file = os.path.join(working_directory, "{}.scn.txt".format(filename))

        # Save the data to the input file
        with open(inp_file, "w") as f:
            f.write(smarts_file_contents)

        # Start the process
        try:
            this_process = subprocess.Popen((executable,), stdout=subprocess.PIPE, stdin=subprocess.PIPE,
                                            stderr=subprocess.PIPE, cwd=smarts)
            # We need to tell smarts where to find the input data
            output, error = this_process.communicate(
                input=bytes('N\n"{0}"\n{1}\nY\n'.format(working_directory, filename), "ASCII"))

            error = error.decode('utf8')

            if len(error) > 0 and 'Note' not in error:
                raise RuntimeError(error)

            data = []
            if os.path.exists(scn_file):
                data = np.loadtxt(skipper(scn_file), unpack=True)
                if target_directory is not None:
                    shutil.copy2(scn_file, scn_file.replace(working_directory, target_directory))
            if os.path.exists(inp_file):
                if target_directory is not None:
                    shutil.copy2(inp_file, inp_file.replace(working_directory, target_directory))
            if os.path.exists(ext_file):
                if target_directory is not None:
                    shutil.copy2(ext_file, ext_file.replace(working_directory, target_directory))
            if os.path.exists(out_file):
                if target_directory is not None:
                    shutil.copy2(out_file, out_file.replace(working_directory, target_directory))

        except RuntimeError as err:
            print('ERROR in SMARTS: {}'.format(err))

        except ValueError as err:
            print('ERROR in SMARTS: {}'.format(err))

        except Exception as err:
            print('ERROR in SMARTS: {}'.format(error_msg))
            print('ERROR in SMARTS: {}'.format(err))

    if len(data) == 0:
        raise ValueError('ERROR in SMARTS: Output file is empty. Likely error in the input parameters.')

    return data


def build_smarts_file(**kwargs):
    try:
        smarts_file_contents = ["'{COMNT}' !Card 1"]
        smarts_file_contents.append("{ISPR} !Card 2")
        smarts_file_contents.append([
                                        "{SPR} !Card 2a",
                                        "{SPR} {ALTIT} {HEIGHT} !Card 2a",
                                        "{LATIT} {ALTIT} {HEIGHT} !Card 2a"
                                    ][kwargs["ISPR"]])

        smarts_file_contents.append("{IATMOS} !Card 3")
        smarts_file_contents.append([
                                        "{TAIR} {RH} '{SEASON}' {TDAY} !Card 3a",
                                        "'{ATMOS}' !Card 3a"
                                    ][kwargs["IATMOS"]])

        smarts_file_contents.append("{IH2O} !Card 4")
        if kwargs["IH2O"] == 0:
            smarts_file_contents.append("{W} !Card 4a")

        smarts_file_contents.append("{IO3} !Card 5")
        if kwargs["IO3"] == 0:
            smarts_file_contents.append("{IALT} {AbO3} !Card 5a")

        smarts_file_contents.append("{IGAS} !Card 6")
        if kwargs["IGAS"] == 0:
            smarts_file_contents.append("{ILOAD} !Card 6a")
            if kwargs["ILOAD"] == 0:
                smarts_file_contents.append(
                    "{ApCH2O} {ApCH4} {ApCO} {ApHNO2} {ApHNO3} {ApNO} {ApNO2} {ApNO3} {ApO3} {ApSO2} !Card 6b"
                )

        smarts_file_contents.append("{qCO2} !Card 7")
        smarts_file_contents.append("{ISPCTR} !Card 7a")

        smarts_file_contents.append("'{AEROS}' !Card 8")
        if kwargs["AEROS"] == "USER":
            smarts_file_contents.append("{ALPHA1} {ALPHA2} {OMEGL} {GG} !Card 8a")

        smarts_file_contents.append("{ITURB} !Card 9")
        smarts_file_contents.append([
                                        "{TAU5} !Card 9a",
                                        "{BETA} !Card 9a",
                                        "{BCHUEP} !Card 9a",
                                        "{RANGE} !Card 9a",
                                        "{VISI} !Card 9a",
                                        "{TAU550} !Card 9a",
                                    ][kwargs["ITURB"]])

        smarts_file_contents.append("{IALBDX} !Card 10")
        if kwargs["IALBDX"] == -1:
            smarts_file_contents.append("{RHOX} !Card 10a")
        smarts_file_contents.append("{ITILT} !Card 10b")
        if kwargs["ITILT"] == 1:
            smarts_file_contents.append("{IALBDG} {TILT} {WAZIM} !Card 10c")
            if kwargs["IALBDG"] == 1:
                smarts_file_contents.append("{RHOG} !Card 10d")

        smarts_file_contents.append("{WLMN} {WLMX} {SUNCOR} {SOLARC} !Card 11")

        smarts_file_contents.append("{IPRT} !Card 12")
        if kwargs["IPRT"] >= 1:
            smarts_file_contents.append("{WPMN} {WPMX} {INTVL} !Card 12a")
            if kwargs["IPRT"] == 2 or kwargs["IPRT"] == 3:
                smarts_file_contents.append("{IOTOT} !Card 12b")
                smarts_file_contents.append("{IOUT} !Card 12c")

        smarts_file_contents.append("{ICIRC} !Card 13")
        if kwargs["ICIRC"] == 1:
            smarts_file_contents.append("{SLOPE} {APERT} {LIMIT} !Card 13a")

        smarts_file_contents.append("{ISCAN} !Card 14")
        if kwargs["ISCAN"] == 1:
            smarts_file_contents.append("{IFILT} {WV1} {WV2} {STEP} {FWHM} !Card 14a")

        smarts_file_contents.append("{ILLUM} !Card 15")

        smarts_file_contents.append("{IUV} !Card 16")

        smarts_file_contents.append("{IMASS} !Card 17")
        smarts_file_contents.append([
                                        "{ZENIT} {AZIM} !Card 17a",
                                        "{ELEV} {AZIM} !Card 17a",
                                        "{AMASS} !Card 17a",
                                        "{YEAR} {MONTH} {DAY} {HOUR} {LATIT} {LONGIT} {ZONE} !Card 17a",
                                        "{MONTH} {LATIT} {DSTEP} !Card 17a",
                                    ][kwargs["IMASS"]])

        smarts_file_schema = "\n".join(smarts_file_contents)
        # print (smarts_file_schema)

        smarts_file_complete = smarts_file_schema.format(**kwargs) + "\n"
    except KeyError as err:
        print("The SMARTS options you have selected require additional data, variables are undefined.")
        print(err)
        raise

    return smarts_file_complete


def get_default_smarts_object():
    """  Provides a dictionary with most of the parameters (values of the CARDS) needed by SMARTS to calculate
     a solar spectrum. It can be used as a template to customize for user defined conditions.

    :return:
    """

    # CONSTANT PARAMETERS   ---------------------------------
    # location and general time info
    latitude = 40.4966  # 'LATIT', deg, latitude
    longitude = -3.4620  # 'LONGIT, deg, longitude
    preasure_model = 1  # 'ISPR', surface preasure model set to real data + altitude correction
    altitude = 0.625  # 'ALTIT', km, altitude above sea level
    altura = 0.0  # 'HEIGHT', m, altude over the ground level
    time_zone = 0  # 'ZONE', time zone
    season = 'SUMMER'  # 'SEASON', season of the year. Can be summer or winter only
    albedo = 9  # 'IALBDX', 'IALBDG', Albedo model = 9, Dry clay soil
    solar_position_mode = 3  # 'IMASS', use the location and time to calculate the position of the sun

    # atmospheric conditions
    atmospheric_data = 1  # 'IATMOS', allows to input the correct atmospheric data
    atmosphere_model = 'USSA'  # 'ATMOS', 'US standard spectrum', general atmospheric model, setting all parameters not given as input
    precipitable_water = 0  # 'IH2O', water vapor data given as input
    ozone = 1  # 'IO3', ozone abundance calculated from atmospheric data
    gas_contents = 0  # 'IGAS', gas abundances set to a particular scenario
    gas_scenairo = 2  # 'ILOAD', light pollution scenario
    CO2content = 370  # 'qCO2', CO2 content in ppmv
    extraterrestial = 0  # 'ISPCTR', extraterrestial spectrum
    aerosol = 'S&F_RURAL'  # 'AEROS', aerosol model
    turbidity = 0  # 'ITURB', select turbidity model
    tau500_param = 0.085  # 'TAU5'

    # output info
    print_info = 2  # 'IPRT', print output in spreadsheet format
    total_variables = 1  # 'IOTOT', total number of variables to print
    which_variables = '4'  # 'IOUT', code of the output variables. Set to all tilted irradiances + experimental with FoV
    wavelenght_min = 280  # 'WLMN', 'WPMN', nm
    wavelenght_max = 4004  # 'WLMX', 'WPMX', nm
    wavelenght_step = 0.5  # 'INTVL', nm
    convolute = 1  # 'ISCAN'
    conv_function = 1  # 'IFILT', Gausian
    conv_wl_min = 300  # 'WV1', nm
    conv_wl_max = 3990  # 'WV2', nm
    conv_FWHM = 4  # 'FWHM', nm
    conv_step = 2  # 'STEP', nm
    tilt = 1  # 'ITILT', enable calculations for a tilt surface
    altitude_tilt = -999  # 'TILT', -999 means 'track the Sun'
    azimuth_tilt = -999  # 'WAZIM', -999 means 'track the Sun'
    solar_constant = 1367  # 'SOLARC, W/m2, solar constant
    sun_correction = 1  # 'SUNCOR', distance to the Sun correction faction. It is not used

    # Colimator information
    circumsolar = 1  # 'ICIRC', if circumsolar radiation is to be calculated
    slope = 1  # 'SLOPE', configuration of the colimator
    aperture = 2.5  # 'APERT'
    limit = 4  # 'LIMIT'

    # Others
    illuminance = 0  # 'ILLUM'
    UVcalc = 0  # 'IUV'
    air_mass = 3  # 'IMASS', set to calculate the air mass from the time and location

    # VARIABLES     ---------------------------------
    comment = 'Test'  # 'COMNT', single line with a comment on the run. Max 64 characters, no spaces but underscores
    P = 940  # 'SPR', mb, surface preasure
    T_air = 17  # 'TAIR', deg, temperature of the air
    T_day = 25  # 'TDAY', deg, average daily temperature
    humid = 30  # 'RH', %, relative humidity
    water_vapour = 1  # 'W', cm, precipitable water
    targetTime = datetime(2015, 5, 19, 12,
                          30)  # has to be split in 'YEAR', 'MONTH', 'DAY', 'HOUR', the later in Local Standard Time

    smarts_input = {
        'LATIT': latitude,
        'LONGIT': longitude,
        'ISPR': preasure_model,
        'ALTIT': altitude,
        'HEIGHT': altura,
        'ZONE': time_zone,
        'SEASON': season,
        'IALBDX': albedo,
        'IALBDG': albedo,
        'IMASS': solar_position_mode,
        'IATMOS': atmospheric_data,
        'ATMOS': atmosphere_model,
        'IH2O': precipitable_water,
        'IO3': ozone,
        'IGAS': gas_contents,
        'ILOAD': gas_scenairo,
        'qCO2': CO2content,
        'ISPCTR': extraterrestial,
        'AEROS': aerosol,
        'ITURB': turbidity,
        'TAU5': tau500_param,
        'IPRT': print_info,
        'IOTOT': total_variables,
        'IOUT': which_variables,
        'WLMN': wavelenght_min,
        'WPMN': wavelenght_min,
        'WLMX': wavelenght_max,
        'WPMX': wavelenght_max,
        'INTVL': wavelenght_step,
        'ISCAN': convolute,
        'IFILT': conv_function,
        'WV1': conv_wl_min,
        'WV2': conv_wl_max,
        'FWHM': conv_FWHM,
        'STEP': conv_step,
        'ITILT': tilt,
        'TILT': altitude_tilt,
        'WAZIM': azimuth_tilt,
        'SOLARC': solar_constant,
        'SUNCOR': sun_correction,
        'ICIRC': circumsolar,
        'SLOPE': slope,
        'APERT': aperture,
        'LIMIT': limit,
        'ILLUM': illuminance,
        'IUV': UVcalc,
        'COMNT': comment,
        'SPR': P,
        'TAIR': T_air,
        'TDAY': T_day,
        'RH': humid,
        'W': water_vapour,
        'YEAR': targetTime.year,
        'MONTH': targetTime.month,
        'DAY': targetTime.day,
        'HOUR': targetTime.hour + (targetTime.minute + targetTime.second / 60) / 60
    }

    return smarts_input


if __name__ == "__main__":
    data = calculate_spectrum_smarts(target_directory='/Users/diego/Downloads')

    import matplotlib.pyplot as plt

    plt.plot(data[0], data[1])
    plt.plot(data[0], data[2])
    plt.plot(data[0], data[3])
    plt.plot(data[0], data[4])
    plt.show()



"""
    Another short python 3.4 travesty by Markus.
    Updated by Diego

"""
from configparser import ConfigParser
from solcore.singleton import Singleton
from typing import Optional


class SourceManagedClassDeprecated(metaclass=Singleton):
    """ Base class that manages the sources of Solcore, from the materials to the unit conversion system.
    It must be defined as a derived class from Singleton so only one instance of this class containing all sources
    exists at any time.
    """

    def __init__(self):
        """ Constructor of the class. """
        self.sources = {}
        self.initialise_cache()

    def add_source(self, name, path):
        """ Read a materials database file in ConfigParser format with case-sensitive options. """
        self.sources[name] = path
        self.read(name)

    def remove_source(self, name, reread=True):
        """ Removes a source from the database. """
        del self.sources[name]
        if reread:
            self.read()

    def read(self, name=None):
        """ Reads the sources and add its contents to the database. """
        if name is None:
            self.initialise_cache()
            for sourcename in self.sources.keys():
                self.database.read(self.sources[sourcename], encoding='utf-8')
        else:
            self.database.read(self.sources[name], encoding='utf-8')

    def initialise_cache(self):
        """ Initialises th database, kept in a ConfigParser format. Quite convenient if you think about it. """
        self.database = ConfigParser()

        # Normally the config parser converts options to lower case; this preserves case.
        self.database.optionxform = str


class SourceManagedClass(metaclass=Singleton):
    """ Base class that manages the sources of Solcore. """

    def __init__(self, sources: dict):
        """ Constructor of the class. """
        self.sources = sources
        self.database = ConfigParser()
        self.database.optionxform = str
        self.database.read(self.sources.values(), encoding='utf-8')

    def read(self, source: Optional[str] = None, value: Optional[str] = None):
        """ Updates the information in the database. """
        if source is not None and value is not None:
            self.sources[source] = value
            self.database.read(value, encoding='utf-8')
        elif source in self.sources and value is None:
            del self.sources[source]
            self.database.read(self.sources.values(), encoding='utf-8')
        else:
            self.database.read(self.sources.values(), encoding='utf-8')



from collections import defaultdict
from solcore import ParameterSystem
import solcore


class Structure(list):
    """ Subclass of lists that stores information of a 'sample' consisting of several 'layers'."""

    def __init__(self, *args, **kwargs):
        super(Structure, self).__init__(*args)
        self.__dict__.update(kwargs)
        self.labels = [None] * len(self)

    def append(self, new_layer, layer_label=None, repeats=1):
        # Pass the arguments to the superclass for extending
        for i in range(repeats):
            # Extend the structure labels
            self.labels.append(layer_label)
            super(Structure, self).append(new_layer)

    def append_multiple(self, layers, layer_labels=None, repeats=1):

        assert type(layers) == type([]), "`append_multiple` only accepts lists for the first argument."

        if layer_labels is not None:
            assert len(layers) == len(
                layer_labels), "When using `layer_labels` keyword a label must be specified for each layer added i.e. layers and layer_labels must have the same number of elements.  Either fix this or simply do not assign any labels (i.e. layer_labels=None)."

        for i in range(repeats):
            # Extend the structure by appending layers
            self.extend(layers)

            # Extend the structure labels by appending an equal number of None values
            # or by appending the actual labels.
            if layer_labels is None:
                self.labels.extend([None] * len(layers))
            else:
                self.labels.extend(layer_labels)

    def __str__(self):

        layer_info = ["  {} {}".format(
            layer,
            self.labels[i] if self.labels[i] is not None else "",
        ) for i, (layer, label) in enumerate(zip(self, self.labels))]

        return "<Structure object\n{}\n{}>".format(str(self.__dict__), "\n".join(layer_info))

    def width(self):
        return sum([layer.width for layer in self])

    def relative_widths(self):
        total = 0
        aggregate_widths = defaultdict(float)
        for layer, comment in zip(self, self.labels):
            aggregate_widths[comment] += layer.width
            total += layer.width

        for layername in aggregate_widths.keys():
            aggregate_widths[layername] /= total

        return aggregate_widths


class Layer:
    """ Class that stores the information about layers of materials, such as thickness and composition.
    It is the building block of the 'Structures' """

    def __init__(self, width, material, role=None, geometry=None, **kwargs):
        """ Layer class constructor.

        :param width: Width of the layer, in SI units.
        :param material: Solcore material
        :param role: Role of the layer.
        :param kwargs: Any other keyword parameter which will become part of the layer attributes
        """
        self.width = width
        self.material = material
        self.role = role
        self.geometry = geometry
        self.__dict__.update(**kwargs)

    def __str__(self):
        """ Representation of the Layer object
        :return: A string with a summary of the layer properties
        """
        widthstring = "{:.3}nm".format(self.width * 1e9)
        return "<{}layer {} {}>".format(
            self.role + " " if self.role != None else "",
            widthstring,
            self.material,
        )


class Junction(list):
    """ Class that groups together the layers that make a junction. Esentially, it is just a list with attributes you can update.
    """

    def __init__(self, *args, **kwargs):
        self.__dict__.update(kwargs)
        list.__init__(self, *args)

    def __str__(self):
        layer_info = ["{}".format(layer) for layer in self]
        return "<Junction object \n\t{}\n\t{}>".format(str(self.__dict__), "\n\t".join(layer_info))


class TunnelJunction(Junction):
    """ Class that contains the minimum definitions for a tunnel junction, ie. a series resistance in our case.
    """
    def __init__(self, *args, **kwargs):
        Junction.__init__(self, *args, **kwargs)

        self.R = kwargs['R'] if 'R' in kwargs.keys() else 1e-16
        self.pn = kwargs['pn'] if 'pn' in kwargs.keys() else True


# CONVERSION UTILITIES
# The following functions are used to move back and forth between the Solcore structures and the Device structures used
# in the PDD solver
def InLineComposition(layer):
    """ Hack to use the Adachi-alfa calculator, provinding the composition as a single string

    :param layer: A layer as defined in the Device structures of the PDD solver
    :return: A mterial string
    """
    comp = layer['properties']['composition']
    if 'element' in comp.keys():
        return comp['material'].replace(comp['element'], comp['element'] + str(comp['fraction']))
    else:
        return comp['material']


def SolcoreMaterialToStr(material_input):
    """ Translate a solcore material into something than can be easily stored in a file and read

    :param material_input: A solcore material
    :return: A dictionary with the name, consituents and composition of the material
    """
    material_string = material_input.__str__().strip('<>').split(" ")
    material_name = material_string[0].strip("'")
    composition = {'material': material_name}

    alloy = True if len(material_input.composition) > 0 else False

    if alloy:
        material_composition = material_string[2].split("=")
        for i, comp in enumerate(material_composition):
            if comp in material_name:
                composition['element'] = material_composition[i]
                composition['fraction'] = float(material_composition[i + 1])

    return composition


def ToSolcoreMaterial(comp, T, execute=False, **kwargs):
    """ It provides a solcore material out of its string composition. The output can be a string with the command or
    a solcore material itself.

    :param comp: A Solcore material as a string
    :param T: The temperature
    :param execute: If a Solcore material must be provided or just a string that would do that.
    :return: A Solcore material or a string with the command to calculate the solcore material
    """
    # It provides a solcore material out of its string composition. The output can be a string with the comand or a solcore material itself.
    if 'element' in comp.keys():
        # A ternary material
        out = 'solcore.material("%s")(T=%s, %s=%s' % (comp['material'], T, comp['element'], comp['fraction'])
    else:
        # A binary material
        out = 'solcore.material("%s")(T=%s' % (comp['material'], T)

    for key in kwargs.keys():
        out = out + ', {}={}'.format(key, kwargs[key])

    out = out + ') '

    if execute:
        return eval(out)
    else:
        return out


def ToLayer(width, material, role):
    """ Creates a Layer based on strings containing the width, the material and the role

    :param width: Width of the layer
    :param material: Material of the layer, as a string
    :param role: The role of the layer
    :return: A Solcore Layer
    """
    return eval('Layer( width=%s, material=%s, role="%s"  )' % (width, material, role))


def ToStructure(device):
    """ Translate the device object (as used by the PDD solver) into a list of solcore Structure that can be used in
    other plugings. Only translate the composition, for now. It works only on non-nested device structures
    (QW, for example, but not in devices with QW)

    :param device: A solcore device structure as used by the PDD solver
    :return: A Solcore Structure
    """

    LayersList = []
    MatList = []

    for i in range(device['numlayers']):
        layer = device['layers'][i]
        MatList.append(ToSolcoreMaterial(layer['properties']['composition'], device['T']))
        LayersList.append(ToLayer(layer['properties']['width'], MatList[i], layer['label']))
        LayersList[-1].material.strained = 'True'

    LayersList = Structure(LayersList)
    LayersList.substrate = ToSolcoreMaterial(device['substrate'], device['T'], execute=True)

    return LayersList



from glob import glob
from pathlib import Path
from pytest import fixture, mark, skip
import os
import subprocess
from .conftest import patch_plots

cwd = os.getcwd()


@fixture(scope="session")
def examples_directory(tmp_path_factory):
    from shutil import copytree

    exdir = Path(__file__).parent.parent / "examples"
    target = tmp_path_factory.getbasetemp() / "examples"
    copytree(exdir, target)
    return target


def example_scripts():
    exdir = Path(__file__).parent.parent / "examples"
    scripts = sorted([os.path.basename(f) for f in glob(str(exdir / "*.py"))])
    return scripts


def example_notebooks():
    exdir = Path(__file__).parent.parent / "examples" / "notebooks"
    scripts = sorted([os.path.basename(f) for f in glob(str(exdir / "*.ipynb"))])
    return scripts


@patch_plots
@mark.parametrize("example", example_scripts())
def test_example_scripts(example, examples_directory):

    env = os.environ.copy()
    env["PYTHONWARNINGS"] = "ignore"
    with open(examples_directory / example) as fp:
        script = "import matplotlib\nmatplotlib.use('Agg')\n" + fp.read()
        process = subprocess.run(
            ["python", "-c", script],
            capture_output=True,
            cwd=str(examples_directory),
            env=env,
        )

    if b"SpiceSolverError" in process.stderr:
        skip("No SPICE solver found.")
    elif b"SmartsSolverError" in process.stderr:
        skip("No SMARTS solver found.")
    elif b"RCWASolverError" in process.stderr:
        skip("No RCWA solver found.")
    elif process.stderr != b"":
        raise Exception(process.stderr.decode())


@patch_plots
@mark.parametrize("example", example_notebooks())
def test_example_notebooks(example, examples_directory):
    import nbconvert
    import nbformat

    exporter = nbconvert.PythonExporter()
    notebooks_folder = examples_directory / "notebooks"
    env = os.environ.copy()
    env["PYTHONWARNINGS"] = "ignore"
    with open(notebooks_folder / example) as fp:
        nbcode = nbformat.reads(fp.read(), as_version=4)
        (code, _) = exporter.from_notebook_node(nbcode)
        script = "import matplotlib\nmatplotlib.use('Agg')\n" + code
        process = subprocess.run(
            ["python", "-c", script],
            capture_output=True,
            cwd=str(notebooks_folder),
            env=env,
        )

    if b"SpiceSolverError" in process.stderr:
        skip("No SPICE solver found.")
    elif b"SmartsSolverError" in process.stderr:
        skip("No SMARTS solver found.")
    elif b"RCWASolverError" in process.stderr:
        skip("No RCWA solver found.")
    elif process.stderr != b"":
        raise Exception(process.stderr.decode())
